      Program StokesSolution

c   ==================================================================================  c
c                                                                                       c
c   This program is for the solution of generalised newtonian fluids of                 c
c   incompressible fluids governed by Stokes equations. The solution is                 c
c   obtained via the weighted residual galerkin finite element method in                c
c   conjunction with the use of 8 nodes isoparametric hexahedral elements.              c
c                                                                                       c
c   The entire computational domain is assumed to be isothermal.                        c
c                                                                                       c
c   Two numerical schemes are developed for the problem solutions:                      c
c   The U-V-W-P method in which velocity components and pressure are regarded           c
c   as primitive variables and discretized as unknowns.                                 c 
c                                                                                       c
c   The second scheme is a modification of the U-V-W-P method which reach convergence   c
c   faster than the traditional U-V-W-P method.                                         c
c                                                                                       c
c   During the program running after typing in some basic data, the user will prompt    c
c   to select with scheme to use.                                                       c
c                                                                                       c    
c   The apparent viscosity is calculated and updated using the power law model          c
c                                                                                       c
c   The system of algebraic equations obtained after the discretization process is      c
c   solved by frontal method.                                                           c 
c                                                                                       c
c   A complete list of options is given on the program listing.                         c
c                                                                                       c
c   The program consists of a main module and subroutines and among the are two         c
c   output subroutines written for Tecplot and Cosmos Geostar for visualizations.       c          
c                                                                                       c
c   The program is written in FORTRAN programming language                              c
c                                                                                       c
c   Developed (not from scratch) by N. Rupert. Boukanga (last updated April 2010)       c
c                                                                                       c
c   The author thanks Prof V. Nassehi for his hard work by originally developing the    c
c   the majority of the subroutines                                                     c
c                                                                                       c
c                                                                                       c
c   ==================================================================================  c
c                                                                                       c
c   work files                                                                          c
c   ==========                                                                          c
c     unit      contents                                                                c 
c ====================================================================================  c
c       51    i    input data file                                                      c
c             i                                                                         c
c       60    i    output file for documentation                                        c
c             i                                                                         c
c       11    i    output file containing velocity field data for                       c
c             i    plotting                                                             c
c             i                                                                         c
c       12    i    output file containing concentration data for                        c
c             i    contour plotting                                                     c 
c             i                                                                         c 
c       14    i    used as a work file in the solver routine                            c   
c             i                                                                         c
c       15    i    stores shape functions and their derivatives at                      c 
c             i    'full' integration points                                            c
c             i                                                                         c
c       17    i    output file containing pressure data for                             c
c             i    contour plotting                                                     c
c             i                                                                         c
c       20    i    output file containing elemental stiffness matrix                    c
c             i    for element number 14 as seen on the mesh                            c
c             i                                                                         c
c       610   i    output for Cosmos Geostar post-processing                            c
c       614   i    output for Tecplot post-processing                                   c
c                                                                                       c 
c ===================================================================================== c  
c                                                                                       c
c     List of variables                                                                 c
c     =================                                                                 c 
c     aa   (   27,  27)   element coefficient matrices on LHS                           c
c     K    (   27,  27)   element coefficient matrices on RHS                           c 
c     b    (    3,  20)   global derivatives of shape functions                         c
c     bc   (maxdf     )   nodal constraints (boundary conditions)                       c
c     conc (maxnp     )   nodal concentrations                                          c
c     cord (maxnp,ndim)   nodal coordinates                                             c   
c     del  (    3,  20)   local  derivatives of shape functions                         c
c     vel  (maxdf     )   nodal velocities  (displacements)                             c
c     dsc1, dsc2          depths of slip layers                                         c   
c     grav1               first  component of the applied body force                    c
c     grav2               second component of the applied body force                    c
c     icord               indicates whether the coordinate system is cartesian (planar) c
c                         or cylindrical (axisymmetric)                                 c
c     tolc                convergence tolerance factor for concentration                c
c     tolp                convergence tolerance factor for pressures                    c
c     tolv                convergence tolerance factor for velocities                   c
c     nbc                 total number of boundary-node constraints                     c
c     ncn                 number of nodes per element                                   c
c     ndf                 degree of freedom per node                                    c
c     ndim                dimensions of the solution domain                             c
c     nel                 total number of elements                                      c    
c     ngaus               number of integration points                                  c  
c     nnp                 total number of nodal points                                  c
c     node (maxel,maxst)   element connectivity                                         c
c     nter                maximum number of iterations for non-newtonian case           c
c     num                 number of integration points per element                      c 
c     p    (    20    )   shape functions                                               c 
c     press(maxnp     )   nodal pressures                                               c
c     r1   (maxdf     )   global load vector  (r.h.s.)                                  c
c     rfrct               friction coefficient (slip)                                   c
c     rr   (   27     )   element load vector                                           c
c     stiff(maxar     )   global stiffness matrix ( a in ax=r.h.s.)                     c
c     rvisc               mu nought;consistency coefficient in power-law model          c
c     power               power law index                                               c
c     stemp               temperature                                                   c  
c     rtem                reference temperature                                         c
c     spress              pressure                                                      c
c     rpress              reference pressure                                            c    
c     tco                 coefficient relating viscosity to temperature                 c
c     pco                 coefficient relating viscosity to pressure                    c
c     gamad               shear rate                                                    c
c     nwr                 no. of sample nodes for recording transient solutions         c
c                                                                                       c
c     ================================================================================= c
c     List of Subroutines                                                               c
c     ===================                                                               c 
c                                                                                       c 
c     bacsub              backsubstitution method for finding the final                 c  
c                         solution vector                                               c
c     clean               cleans the arrays and prepares them for                       c
c                         solution                                                      c
c     conc                calculates the concentrations                                 c
c     contol              makes a check for the convergence                             c
c     deriv               calculates the jacobian matrix, its determinant               c
c                         and	global derivatives of the shape functions               c
c     flow Stokes         calculates the velocities and pressures via the UVWP method   c
c     flow Stokes2        calculates the velocities and pressures via the modified      c
c                         UVWP method                                                   c
c     front               frontal method for solving the final set                      c
c                         of equations                                                  c
c     gaussp              specifies the gauss points and weights for                    c
c                         quadrature integration                                        c 
c     getbcd              specifies the primary boundary conditions                     c
c     getelm              specifies the nodal connectivity array                        c
c     getmat              reads the input material data                                 c
c     getnod              reads the nodal co-ordinates for cartesian                    c
c                         and axisymmetric systems                                      c
c     lumpm               evaluates the terms of the mass matrix                        c
c     output              prints the final solution                                     c
c     putbcv              imposes the primary boundary conditions for                   c  
c                         velocity                                                      c
c     putbcc              imposes the primary boundary conditions for                   c
c                         concentration                                                 c
c     setprm              Sets the location data for nodal degrees of                   c
c                         freedom                                                       c
c     shape               calculates the shape functions and their                      c
c                         derivatives                                                   c
c     slip                identifies the upper and lower boundary                       c
c                         layers.                                                       c 
c     stress              calculates stress components at integration                   c
c                         points                                                        c
c     visca               calculates the viscosity                                      c
c     viscb               calculates virtual viscosity for slip walls                   c  
c                                                                                       c
c    ================================================================================== c 
      parameter (maxel  = 60000    )
	parameter (maxnp  = 37000    )
	parameter (maxbc  = 20000    )
      parameter (maxdf  = maxnp*4  )
	parameter (maxst  = 80      )
	parameter (maxfr  = 5000   )
      parameter (ndim   = 3       )

      implicit real*8 (a-h,o-z)
c
c    Storage allocation
c    ==================
c      
	dimension title (         80)
      dimension node  (maxel,maxst) ,pmat (maxel,  9) ,cord (maxnp,ndim)
      dimension ncod  (maxdf      ), bc   (maxdf    ) 
      dimension ibc   (maxbc      ) ,jbc  (maxbc    ) ,vbc  (maxbc    )
      dimension vel   (maxdf      ) ,conc (maxnp    ) ,press(maxnp    )
	dimension r1    (maxdf      ) 
      dimension clump (maxnp      ) ,stres(maxnp,  6) 
      dimension vet   (maxdf      ) ,cet  (maxnp    ) ,pet  (maxnp    ) 
      dimension nopp  (maxdf      )
      dimension aa    (maxst,maxst) ,rr   (maxst    )
      dimension xg    (          3) ,cg   (        3)
      dimension p     (         20) ,del  (    3, 20) ,b    (   3,  20)
      dimension ldest (maxst      ) ,kdest(maxst    ) ,nk   (maxst    )
      dimension eq    (maxfr,maxfr) ,lhed (maxfr    ) ,khed (maxfr    )
      dimension kpiv  (maxfr      ) ,lpiv (maxfr    ) ,jmod (maxfr    )
      dimension qq    (maxfr      ) ,pvkol(maxfr    ) ,sinv (maxel, 27)
      dimension mdf   (maxdf      ) ,ndn  (maxdf    ) 
	dimension ldsc  (22         )
      dimension temp  (maxnp      ) ,actpress(maxnp)
      dimension rmat1 (maxel,   13) ,rmat2(maxel, 13)
      character *20 filnam
c
c    Opening of input and output data files 
c    ======================================
c      
      call GFMFEM
      
	print*,'enter the name of your data file'
	read(*,2000) filnam	

      open(unit=51,File=filnam,access='sequential',form='formatted',
     1     status="unknown",iostat=ios)
           
	open(unit=60,file='res.txt',access='sequential',form='formatted',
     1     status="unknown",iostat=ios)
	 
	open(unit=17,file='stress.txt',access='sequential',form='formatted',
     1     status="unknown",iostat=ios)
      open(unit=20,file='stiffmat',access='sequential',form='formatted',
     1     status="unknown",iostat=ios)
           
		
      open(unit=14,form='unformatted',status='scratch',iostat=ios)
      open(unit=15,form='unformatted',status='scratch',iostat=ios)

        if(ios==0)then
			print*,"files opened"
	  else 
			print*,"files not opened"
	  stop  
	  end if

	rewind 51
	rewind 60
	rewind 20
c
c     Initialize arrays
c    =================

      do 5010 itl = 1,maxel
      do 5010 ivl = 1,80
                          node (itl,ivl) = 0
5010  continue
      do 5020 itl = 1,maxel
      do 5020 ivl = 1,8
                          pmat (itl,ivl) = 0.0
5020  continue
      do 5030 itl = 1,maxnp
      do 5030 ivl = 1,3
                          cord (itl,ivl) = 0.0
5030  continue
      do 5040 itl = 1,maxnp
      do 5040 ivl = 1,6
                          stres(itl,ivl) = 0.0
5040  continue
      do 5050 itl = 1,maxdf
	                    vel  (itl    ) = 0.0 
5050  continue     
      do 5060 itl = 1,maxdf
                          ncod   (itl)   = 0
                          r1     (itl)   = 0.0
                          bc     (itl)   = 0.0
                          vet    (itl)   = 0.0
                          mdf    (itl)   = 0
                          ndn    (itl)   = 0
                          nopp   (itl)   = 0
5060  continue
      do 5070 itl = 1,maxnp
                          clump  (itl)   = 0.0
                          cet    (itl)   = 0.0
                          conc   (itl)   = 0.0
						pet    (itl)   = 0.0
						press  (itl)   = 0.0
5070  continue
      do 5080 itl = 1,maxbc
                          ibc   (itl)    = 0
                          jbc   (itl)    = 0
                          vbc   (itl)    = 0.0
5080  continue
      do 5090 itl = 1,20
                          del   (1,itl)  = 0.0
                          del   (2,itl)  = 0.0
                          del   (3,itl)  = 0.0
5090  continue
      do 5100 itl = 1,maxst
                          ldest (itl  )  = 0
                          kdest (itl  )  = 0
                          nk    (itl  )  = 0
5100  continue
      do 5110 itl = 1,maxfr
                          lhed  (itl  )  = 0
                          khed  (itl  )  = 0
                          kpiv  (itl  )  = 0
                          lpiv  (itl  )  = 0
                          jmod  (itl  )  = 0
                          qq    (itl  )  = 0.0
                          pvkol (itl  )  = 0.0
      do 5110 ill = 1,maxfr
                          eq    (itl,ill)= 0.0
5110  continue
c-----------------------------------------------------------------------
    
c     Title of the program                
c     ====================

	if(.not. eof(51)) read (51,2010) title
      write(60,4010) title
	
c     Element description data
c     ========================
     
      if (.not. eof(51)) read (51,2020) ncn ,ngaus
	print*, "ncn, ngaus read"
      write(60,4020) ncn ,ngaus

c     Mesh, boundary condition and material parameters
c     ================================================
	
	if (.not. eof(51)) read (51,2030) nnp  ,nel  ,nbc  ,nmat
	print*, "nnp, nel ,nbc ,nmat read"
      if (.not. eof(51)) read (51,2040) ntep ,icord
c
      if(icord.eq.0) write(60,4030)
      if(icord.eq.1) write(60,4040)

      write(60,4050)
     
      if(ntep.eq.0) ntep=1

c    ===========================================================
c    icord=0 coordinate system is cartesian   (   planar   )
c    icord=1 coordinate system is cylindrical (axisymmetric)
c
c    if ntep = 1 then computed result after every iteration will
c    be printed ;if you do not need the result of intermediate
c    computations choose your own ntep;the result of first and
c    converged solutions will always be printed.
c    ===========================================================

                if(nnp   .eq.0  .or.nnp  .gt.maxnp) then
                                             write(60,4060)
            elseif(nel   .eq.0  .or.nel  .gt.maxel) then
                                             write(60,4060)
            elseif(nbc   .eq.0  .or.nbc  .gt.maxbc) then
                                             write(60,4060)
            elseif(nmat  .eq.0  .or.nmat .gt.maxel) then
                                             write(60,4060)
            
		  print*, "the program is aborted"
            stop
	      
		      endif
      
	write(60,4070) nnp ,nel ,nbc ,nmat

      if (.not. eof(51)) read (51,2050) grav1, grav2, grav3
	print*, "grav1 grav2 grav3 read"
      write(60,4080) grav1, grav2, grav3

      if (.not. eof(51)) read(51,2060) tolv ,tolp, tolc
      print*, "tolv, tolp, tolc read"

      maxer=maxel

c    ===========================================================================
c    Read input data from main data file and prepare arrays for solution process
c    ===========================================================================

       call getmat(nel,nmat,pmat,51,60,maxel,rtem,rpef)
       call getnod(nnp,cord,51,60,maxnp,ndim,icord)
       call getelm(nel,ncn,node,51,60,maxer)
       call getbcd(nbc,ibc,jbc,vbc,51,60,maxbc)
	

c    ==========================================================================
c						 Start of the time loop
c    ==========================================================================
c
c     Set control parameters (default values are overwritten by input data
c     if specified)
c
c     ncn      number of nodes per element
c     ngaus    number of integration points
c     nter     maximum number of iterations for non-newtonian case
c     ndim     number of space dimensions in the solution domain
c
c      nter  = 5
       num = 13
	                                do 5125 iel = 1,maxel
	                                  do 5125 lg = 1,num
	                                  sinv(iel,lg)=0.0
5125                                  continue 
                        
                                      do 5130 ivel= 1,maxdf
                                         vel (ivel)  =   0.0
5130                                  continue

                                      do 5140 item= 1,maxnp
                                         temp(item)  =    rtem
5140                                  continue

c    ==============================================================================
c                                 Transient data
c    ==============================================================================
c
c     stime		starting time
c
c     deltat		time increment
c
c     alpha       indicates the choice of method being employed in alpha
c	            time stepping technique (backward difference, 
c				forward difference, central difference, galerkin) 
c
c     nter        maximum number of time steps being employed for finding solution
c     ============================================================================

      print*,"  "
      
      print*,"Enter the number of time steps desired"
c
      read*, nter	

      write(60,4100) nter

      print*,"Enter the delta t desired"
c
      read*, deltat	
 
      write(60,4110) deltat       


1111 	print*," Enter the value of alpha: any number between 0 and 1 "
	    
	read*, alpha
	    
	if(alpha < 0 .OR. alpha > 1) then
	 
	print*, " Invalid alpha value entered, type in another value" 
	
	goto 1111  
	
	end if 

      print*, "alpha=",alpha

        tcode = 0
        
	  Print *,'           '
	  
	   print*," Select a scheme "
	  

	  Print *,'1: UVW-P scheme with Taylor-Galerkin method'
	  Print *,'2: Modified UVW-P scheme with Taylor-Galerkin method'
	  
	  read(*,*) tcode


	
      do 5150 iter = 1 ,nter
      print*,'iter=',iter
	time = iter*deltat

      write(60,4090) iter

c    ====================================================================
c                         Calculate Nodal Velocities & Pressures
c    ====================================================================
      icho=1
                             rewind 11
                             rewind 14
                             rewind 15
                             ndf   = 4
                             ntov  = ndf * nnp
                             ntrix = ndf * ncn
      call clean
     1    (ncn  ,nel  ,ndf  ,node ,r1   ,maxel,maxst,maxdf,
     2     bc   ,ncod ,icho )

      call setprm
     1    (nnp  ,nel  ,ncn  ,node ,ndf  ,maxel,maxst,ndn  ,ntrix,
     2     maxdf,ntov ,mdf  ,nopp )

      call putbcv
     1  (nnp ,nbc ,ibc ,jbc ,vbc ,ncod ,bc,maxbc,maxdf,maxel,maxst,
     2   node)


c     idv4 is the file specifier for unit=20
c     ======================================
 
      idv4=20			 

      
        do 5160 iel=1,nel


     
        if (tcode ==1) then
       
       call flowStokes(node ,cord ,pmat ,nopp ,mdf ,ndn ,ncod ,
	1bc ,vel  ,press, r1, temp,ldest,kdest,nk ,eq ,lhed ,khed ,kpiv ,lpiv,
     2jmod, qq, pvkol, iter ,nel ,ncn , ngaus,grav1,
     3grav2, grav3, p, del, b, ntrix, maxel, maxnp, maxst, maxfr, maxdf,
     4ndim ,aa ,xg ,da ,ntov ,num, icord, rr, iel, del1,deltat,alpha,
     5idv4,sinv, icho, nnp, tref)
     
     
        else if (tcode ==2) then
        
      call flowStokes2(node ,cord ,pmat ,nopp ,mdf ,ndn ,ncod ,
	1bc ,vel  ,press, r1, temp,ldest,kdest,nk ,eq ,lhed ,khed ,kpiv ,lpiv,
     2jmod, qq, pvkol, iter ,nel ,ncn , ngaus,grav1,
     3grav2, grav3, p, del, b, ntrix, maxel, maxnp, maxst, maxfr, maxdf,
     4ndim ,aa ,xg ,da ,ntov ,num, icord, rr, iel, del1,deltat,alpha,
     5idv4,sinv, icho, nnp, tref)

       else 
	 stop
	 endif 


5160  end do

c     =================================================================================
c              calculates the second invariant of rate of deformation
c                         tensor at integration points.
c     =================================================================================
      call secinv
     1   (nel  ,nnp  ,ncn  ,ngaus,node ,sinv ,cord ,p  ,b,
     2    del  ,da   ,vel  ,maxnp,maxel,maxst,ndim ,icord,
     3    maxdf,num)

c     ================================================================
c                           Convergence check
c     ================================================================
      call contol(vel ,conc ,iter ,ntov ,nnp,maxnp,maxdf,errov, errop
	1,vet ,cet, pet, press)  

c     ================================================================
c *** calculation of the nodal stress;using variational recovery
c     ================================================================
      call lumpm
     1    (clump,nnp  ,maxnp,nel  ,ngaus,p   ,del   ,b  ,maxst,
     2     node ,maxel,ncn  )
c
      call stress
     1     (nel,nnp,ncn ,node ,p , b , da ,vel  ,maxnp, maxel,  maxst , 
     2      maxdf, stres, press, rvisc ,clump ,ngaus  ) 


c	================
c     Print the output
c     ================

      iiter=(iter/ntep)*ntep

      if(iter.eq.1.or.iiter.eq.iter) then
      
	call output
	1   (nnp  ,vel  ,press, maxdf,maxnp,icord, stres)
	
	end if 
	
	if(iter.eq.nter) then
	
	
	call cosmos
	1       (nnp  ,  vel  ,  press   ,  maxdf  ,  maxnp  , icord   ,
     2        pmat ,  maxel,  actpress, nel                         )
     
     

      call tecplot
	1       (nnp  ,  vel  ,  press   ,  maxdf  ,  maxnp  , icord   ,
     2        pmat ,  maxel,  actpress,  cord   ,  ncn    , nel     ,
     3        node ,  ndim  )
     
      
     
	endif
c
c
c     =================================================================
c                            End of time loop
c     =================================================================

5150  continue

      close(51)
      close(unit=60)
      close(unit=11)
      close(unit=14)
      close(unit=15)
      close(20)
c
c    ===================================================================
c                         Read statements
c    ===================================================================
c
c
2000  format(a)
2010  format(80a)
2020  format(2i5)
2030  format(4i5)
2040  format(2i5)
2050  format(3f10.0)
2060  format(3f10.5)
c
c
c    ===================================================================
c                         Write statements
c    ===================================================================

4010  format(' ',5(/),' ',20x,60('*'),/' ',20x,'*',58x,'*',/
     1' ',20x,'*','  A  three  dimensional finite element model of a  ',
     27x,'*',/' ',20x,'*','  non-newtonian isothermal flow using ',
     320x,'*',/' ',20x,'*','  the UVWP method.  ',38x,'*',/' ',20x,'*',
     558x,'*',/' ',20x,60('*')///,' ',20x,80('-'),/' ',20x,80a,/' ',
     620x,80('-'),///)
c
4020  format(' ',20x,3('['),' element description data',10('.'),/
     125x,'no.of nodes per element                  =',i10,/
     225x,'no.of integration points                 =',i10,/
     3//)
c
4030  format('    *** coordinate system is cartesian (planar) ***')
4040  format('*** coordinate system is cylindrical(axisymmetric) ***')
4050  format(' ')
4060  format(' ',10('['),'input data unacceptable',10(']')///)
c
4070  format(' ',20x,3('['),' mesh description data ',10('.'),/
     125x,'no.of nodal points                       =',i10,/
     225x,'no.of elements                           =',i10,/
     325x,'no.of nodal constraints on boundary      =',i10,/
     425x,'no.of different materials                =',i10,//)
c
4080  format(' ',20x,3('['),' uniform body force vector ',10('.'),/
     125x,'grav1                                    =',f15.4,/
     225x,'grav2                                    =',f15.4,/
     325x,'grav3                                    =',f15.4,//)
4090  format(///'     iteration no.',i5,//)
4100  format(///' Total number of time steps  =',i5,//)
4110  format(///' Deltat                      =',f15.4,//)

c
c     ==========================================================================
      end program 
c     ==========================================================================
c
      subroutine gaussp(ngaus,xg,cg)
c
      implicit double precision(a-h,o-z)
c
c     x(g)   specifies the coordinates of the Gauss points
c     c(g)   specifies the Gauss weights
c
      dimension xg(3),cg(3)

      if(ngaus.eq.1) then
      xg(1)=0.0
      cg(1)=2.0
      elseif(ngaus.eq.2)  then
      xg(1) = 0.57735026919d00
      xg(2) = -xg(1)
      cg(1) = 1.00
      cg(2) = 1.00
      else
      xg(1) = 0.77459666924d00
      xg(2) = 0.0
      xg(3) = -xg(1)
      cg(1) = 0.55555555556d00
      cg(2) = 0.88888888889d00
      cg(3) = cg(1)
c
      endif
      return
      end
c
c     =================================================================

      subroutine shape ( xi , eta , zeta, p ,del , ncn )
      implicit double precision (a-h,o-z)
c
      DIMENSION p(20) ,del(3,20)
      if (ncn.eq.8) then
         del(1,1)=-0.125*(1-eta)*(1-zeta)
         del(1,2)=-0.125*(1+eta)*(1-zeta)
         del(1,3)= 0.125*(1+eta)*(1-zeta)
         del(1,4)= 0.125*(1-eta)*(1-zeta)
         del(1,5)=-0.125*(1-eta)*(1+zeta)
         del(1,6)=-0.125*(1+eta)*(1+zeta)
         del(1,7)= 0.125*(1+eta)*(1+zeta)
         del(1,8)= 0.125*(1-eta)*(1+zeta)
C.......................................................
         del(2,1)=-0.125*(1-xi)*(1-zeta)
         del(2,2)= 0.125*(1-xi)*(1-zeta)
         del(2,3)= 0.125*(1+xi)*(1-zeta)
         del(2,4)=-0.125*(1+xi)*(1-zeta)
         del(2,5)=-0.125*(1-xi)*(1+zeta)
         del(2,6)= 0.125*(1-xi)*(1+zeta)
         del(2,7)= 0.125*(1+xi)*(1+zeta)
         del(2,8)=-0.125*(1+xi)*(1+zeta)
C...........................................................
         del(3,1)=-0.125*(1-xi)*(1-eta)
         del(3,2)=-0.125*(1-xi)*(1+eta)
         del(3,3)=-0.125*(1+xi)*(1+eta)
         del(3,4)=-0.125*(1+xi)*(1-eta)
         del(3,5)= 0.125*(1-xi)*(1-eta)
         del(3,6)= 0.125*(1-xi)*(1+eta)
         del(3,7)= 0.125*(1+xi)*(1+eta)
         del(3,8)= 0.125*(1+xi)*(1-eta)
C...........................................................

         p(1)=0.125*(1-xi)*(1-eta)*(1-zeta)
         p(2)=0.125*(1-xi)*(1+eta)*(1-zeta)
         p(3)=0.125*(1+xi)*(1+eta)*(1-zeta)
         p(4)=0.125*(1+xi)*(1-eta)*(1-zeta)
         p(5)=0.125*(1-xi)*(1-eta)*(1+zeta)
         p(6)=0.125*(1-xi)*(1+eta)*(1+zeta)
         p(7)=0.125*(1+xi)*(1+eta)*(1+zeta)
         p(8)=0.125*(1+xi)*(1-eta)*(1+zeta)
C...........................................................

      endif
      return
      end
      
c     =================================================================
c
      subroutine deriv
     1    (iel  ,ig   ,jg   ,kg, p    ,del  ,b ,ncn ,da ,cg ,node,
     2     cord ,maxel,maxnp)
c
      implicit double precision(a-h,o-z)
      dimension p(20),b(3,40),del(3,20),cg(3),cj(3,3),cji(3,3)
      dimension node(maxel,27),cord(maxnp,3)
c

	
	
	do 6010 j=1,3
      do 6010 l=1,3
      gash=0.0
            
      do 6020 k=1,ncn
      
      nn=iabs(node(iel,k))
    

6020  gash=gash + del(j,k)*cord(nn,l)
      
	cj(j,l)=gash 

6010  continue

      

      detj =  cj(1,1)*cj(2,2)*cj(3,3)+cj(2,1)*cj(3,2)*cj(1,3)
     1      + cj(1,2)*cj(2,3)*cj(3,1)-cj(1,3)*cj(2,2)*cj(3,1)
     2      - cj(1,2)*cj(2,1)*cj(3,3)-cj(2,3)*cj(3,2)*cj(1,1)   
      
	if(detj.le.0.0) then
      write(60,3010) iel,detj
3010  format(1x ,' Error: Zero or Negative Jacobian. ', i6,g20.5)
      stop 
     
c
      endif

      cji(1,1) =  (cj(2,2)*cj(3,3)-cj(3,2)*cj(2,3))  / detj
      cji(1,2) = ((cj(1,2)*cj(3,3)-cj(3,2)*cj(1,3))) / detj
      cji(1,3) =  (cj(1,2)*cj(2,3)-cj(2,2)*cj(1,3))  / detj
      cji(2,1) = ((cj(2,1)*cj(3,3)-cj(3,1)*cj(2,3))) / detj
      cji(2,2) =  (cj(1,1)*cj(3,3)-cj(3,1)*cj(1,3))  / detj
      cji(2,3) = ((cj(1,1)*cj(2,3)-cj(2,1)*cj(1,3))) / detj
      cji(3,1) =  (cj(2,1)*cj(3,2)-cj(3,1)*cj(2,2))  / detj
      cji(3,2) = ((cj(1,1)*cj(3,2)-cj(3,1)*cj(1,2))) / detj
      cji(3,3) =  (cj(1,1)*cj(2,2)-cj(2,1)*cj(1,2))  / detj

c
      

      do 6030 j=1,3
      do 6030 l=1,ncn
      b(j,l)=0.0
      do 6030 k=1,3
6030  b(j,l) = b(j,l) + cji(j,k) * del(k,l)
      
     	da= detj*cg(ig)*cg(jg)*cg(kg)
c

      return
      end

c
c    =================================================================
c
      subroutine front
     1    (aa   ,rr   ,iel  ,nop  ,maxel,maxst,ldest,kdest,nk   ,maxfr,
     2     eq   ,lhed ,khed ,kpiv ,lpiv ,jmod ,qq   ,pvkol,vel  ,r1   ,
     3     ncod ,bc   ,nopp ,mdf  ,ndn  ,maxdf,nel  ,maxte,ntov ,lcol ,
     4     nell ,ntra, press,icho,c,akf,ak )
c
c     Frontal elimination routine using diagonal pivoting
c
      implicit double precision(a-h,o-z)
      dimension aa   (maxst,maxst) ,rr   (maxst)
      dimension nop  (maxel,maxst)
      dimension ldest(maxst)       ,kdest(maxst)       ,nk   (maxst)
      dimension eq   (maxfr,maxfr) ,lhed (maxfr)       ,khed (maxfr)
      dimension kpiv (maxfr)       ,lpiv (maxfr)
      dimension jmod (maxfr)       ,qq   (maxfr)       ,pvkol(maxfr)
      dimension vel  (maxte)       ,r1   (maxdf)       ,ncod (maxdf)
      dimension bc   (maxdf)       ,nopp (maxdf)       ,mdf  (maxdf)
      dimension ndn  (maxdf)       ,press(maxdf)
c
c     nlp and ndl are the file specifiers for units 60 and 14 respectively
c     ====================================================================

      nlp=60
      nd1=14
c
c     Prefront
c     ========
      nmax=maxfr
      ncrit=20
      nlarg=maxfr-10
      if(iel.eq.1) nell = 0
      if(iel.eq.1) ntra = 1
      if(ntra.eq.0) goto 6040
      nmax = maxfr     
      ntra = 0
      ncrit = 20       
      lfron = 0
      nlarg = nmax-10  
c
c     Find last appeareance of each node
c     ==================================
      nlast = 0
      do 6010 i = 1,ntov
      do 6020 n = 1,nel
      jdn = ndn(n)
      do 6030 l = 1,jdn
      if(nop(n,l).ne.i)go to 6030
      nlast1 = n
      nlast = n
      l1 = l
6030  continue
6020  continue
      if(nlast.eq.0) go to 6010
      nop(nlast,l1) = -nop(nlast,l1)
      nlast = 0
6010  continue
      ntrix = jdn
c
c     Assembly
c     ========   
6040  continue
      if(iel.gt.1) go to 6060
      lcol = 0
      do 6050 i = 1,nmax
      do 6050 j = 1,nmax
      eq(j,i) = 0.
6050  continue
6060  nell = nell+1
      n = nell
      jdn = ndn(nell)
      kc = 0
      do 6070 j = 1,jdn 
      nn = nop(n,j)
      m = iabs(nn)
      k = nopp(m)
      idf = mdf(m)
      r1(m) = rr(j)+r1(m)
      do 6070 l = 1,idf
      kc = kc+1
      ii = k+l-1
      if(nn.lt.0)ii = -ii
      nk(kc) = ii
6070  continue
c
c     Set up heading vectors
c     ======================
c
      do 6080 lk = 1,kc
      node = nk(lk)
      if(lcol.eq.0)goto 6100
      do 6090 l = 1,lcol
      ll = l
      if(iabs(node).eq.iabs(lhed(l)))go to 6110
6090  continue
6100  lcol = lcol+1
      ldest(lk) = lcol
      lhed(lcol) = node
      go to 6080
6110  ldest(lk) = ll
      lhed(ll) = node
6080  continue
      if(lcol.le.nmax)go to 6130
      nerror = 2
      write(nlp,3010)nerror
      stop
6130  continue
      do 6140 l = 1,kc
      ll = ldest(l)
      do 6140 k = 1,kc
      kk = ldest(k)
      eq(kk,ll) = eq(kk,ll)+aa(k,l)
6140  continue
      if(lcol.lt.ncrit.and.nell.lt.nel) return
c
c     Find out which matrix elements are fully assembeled
c     ===================================================
6150  lc = 0
      ir = 0
      do 6160 l = 1,lcol
      kt = lhed(l)
      if(kt.ge.0)go to 6160
      lc = lc+1
      lpiv(lc) = l
      kro = iabs(kt)
      if(ncod(kro).ne.1)go to 6160
      ir = ir+1
      jmod(ir) = l
      ncod(kro) = 2
      r1(kro) = bc(kro)
6160  continue
c
c     Modify equations with applied boundary conditions
c     =================================================
      if(ir.eq.0)go to 6190
      do 6170 irr = 1,ir
      k = jmod(irr)
      kh = iabs(lhed(k))
      do 6180 l = 1,lcol
      eq(k,l) = 0.
      lh = iabs(lhed(l))
      if(lh.eq.kh)eq(k,l) = 1.
6180  continue
6170  continue
6190  continue
      if(lc.gt.0)go to 6200
      ncrit = ncrit+10
c     write(nlp,3020)ncrit
      if(ncrit.le.nlarg) return
      nerror = 3
      write(nlp,3030)nerror
      stop
6200  continue
c
c     Search for absolute pivot
c     =========================
      pivot = 0.
      do 6210 l = 1,lc
      lpivc = lpiv(l)
      kpivr = lpivc
      piva = eq(kpivr,lpivc)
      if(abs(piva).lt.abs(pivot))go to 6220
      pivot = piva
      lpivco = lpivc
      kpivro = kpivr
6220  continue
6210  continue
      if(pivot.eq.0.0) return
c
c     Normalise pivotal row
c     =====================
      lco = iabs(lhed(lpivco))
      kro = lco

c     if(nit.eq.0.or.npra.eq.0)go to 6230
c6230 continue

      if(abs(pivot).lt.0.1d-28) write(nlp,3050)
      do 6240 l = 1,lcol
      qq(l) = eq(kpivro,l)/pivot
6240  continue
      rhs = r1(kro)/pivot
      r1(kro) = rhs
      pvkol(kpivro) = pivot
c
c     Eliminate then delete pivotal row and column
c     ============================================
      if(kpivro.eq.1)go to 6300
      kpivr = kpivro-1
      do 6250 k = 1,kpivr
      krw = iabs(lhed(k))
      fac = eq(k,lpivco)
      pvkol(k) = fac
      if(lpivco.eq.1.or.fac.eq.0.)go to 6270
      lpivc = lpivco-1
      do 6260 l = 1,lpivc
      eq(k,l) = eq(k,l)-fac*qq(l)
6260  continue
6270  if(lpivco.eq.lcol)go to 6290
      lpivc = lpivco+1
      do 6280 l = lpivc,lcol
      eq(k,l-1) = eq(k,l)-fac*qq(l)
6280  continue
6290  r1(krw) = r1(krw)-fac*rhs
6250  continue
6300  if(kpivro.eq.lcol)go to 6360
      kpivr = kpivro+1
      do 6310 k = kpivr,lcol
      krw = iabs(lhed(k))
      fac = eq(k,lpivco)
      pvkol(k) = fac
      if(lpivco.eq.1)go to 6330
      lpivc = lpivco-1
      do 6320 l = 1,lpivc
      eq(k-1,l) = eq(k,l)-fac*qq(l)
6320  continue
6330  if(lpivco.eq.lcol)go to 6350
      lpivc = lpivco+1
      do 6340 l = lpivc,lcol
      eq(k-1,l-1) = eq(k,l)-fac*qq(l)
6340  continue
6350  r1(krw) = r1(krw)-fac*rhs
6310  continue
6360  continue
c
c     Write pivotal equation on disc
c     ==============================   
      write(nd1) kro,lcol,lpivco,(lhed(l),qq(l),l = 1,lcol)
      do 6370 l = 1,lcol
      eq(l,lcol) = 0.
      eq(lcol,l) = 0.
6370  continue
c
c     Rearrange heading vectors
c     =========================
      lcol = lcol-1
      if(lpivco.eq.lcol+1)go to 6390
      do 6380 l = lpivco,lcol
      lhed(l) = lhed(l+1)
6380  continue
6390  continue
c
c     Determine whether to assemble,eliminate,or backsubstitute
c     =========================================================
      if(lcol.gt.ncrit)go to 6150
      if(nell.lt.nel) return
      if(lcol.gt.1)go to 6150
      lco = iabs(lhed(1))
      kpivro = 1
      pivot = eq(1,1)
      kro = lco
      lpivco = 1
      qq(1) = 1.

c     if(nit.eq.0.or.npra.eq.0)go to 6400
c     write(nlp,3040)lco,kro,pivot

      if(abs(pivot).lt.1d-28)go to 6410

c6400 continue

      r1(kro) = r1(kro)/pivot
      write(nd1) kro,lcol,lpivco,lhed(1),qq(1)
c
c     start back-substitution
c     =======================
      call bacsub
     1    (ntov ,ncod ,bc   ,r1   ,vel  ,press, maxfr,qq   ,lhed ,nd1,
     2     icho)
c
c     main exit with solution
c     =======================
6410  continue
c
3010  format(/'  nerror=',i5//
     1 '  the difference nmax-ncrit is not sufficiently large'
     1/'  to permit the assembly of the next element---'
     1/'  either increase nmax or lower ncrit'
     1/)
c3020  format('  frontwidth value=',i4)
3030  format(/'  nerror=',i5//
     1 '  there are no more rows fully summed,this may be due to---'
     1/'  (1)incorrect coding of nop or nk arrays'
     1/'  (2)incorrect value of ncrit. increase ncrit to permit'
     1/'     whole front to be assembled'
     1/)

c3040  format(13h pivotal row=,i4,16h pivotal column=,i4,7h pivot=,e20.10
c    1)

3050  format('  warning-matrix singular or ill conditioned')

      return
      end
c
c     ==================================================================
      subroutine bacsub
     1    (ntotl,ifix ,vfix ,rhs  ,soln ,soln1, mfrnt,rwork,iwork,idv2,
     2     icho )
c
c
      implicit double precision(a-h,o-z)
      dimension ifix (ntotl),vfix (ntotl),rhs (ntotl),soln (ntotl)
      dimension rwork(mfrnt) ,iwork(mfrnt) ,soln1(ntotl) 
c
c
c 
                         do 6010 ipos=1,ntotl
                         soln(ipos) =0.0
                         if(ifix(ipos).ne.0) soln(ipos)=vfix(ipos)
6010                     continue
                         
	                   
 
      do 6020 kpos=1,ntotl
c
      backspace idv2
      read(idv2)     ipos,ifrnt,jfrnt,(iwork(k),rwork(k),k=1,ifrnt)
      backspace idv2
c
      if(ifix(ipos).ne.0)  go to 6020
c
      ww           = 0.0
      rwork(jfrnt) = 0.0
c
                         do 6030 k=1,ifrnt
                         jpos=iabs(iwork(k))
                         ww  =ww - rwork(k)*soln(jpos)
6030                     continue
c
      soln (ipos)=rhs(ipos)+ww
      
6020  continue

	if (icho .eq. 2) goto 6050

      do  6040 ipos  = ((3*ntotl)/4)+1 , ntotl
	       j       = ipos -((3*ntotl)/4)
	      soln1(j) = soln(ipos)

6040  continue 

6050	continue

      return
      end
c
            
c     ==================================================================
c     Stokes Solution based on UVW-P scheme via Taylor-Galerkin
c     time stepping method.
c     ==================================================================

      subroutine flowStokes(node ,cord ,pmat ,nopp ,mdf ,ndn ,ncod ,
	1bc ,vel  ,press, r1, temp,ldest,kdest,nk ,eq ,lhed ,khed ,kpiv ,lpiv,
     2jmod, qq, pvkol, iter ,nel ,ncn , ngaus,grav1,
     3grav2, grav3, p, del, b, ntrix, maxel, maxnp, maxst, maxfr, maxdf,
     4ndim ,aa ,xg ,da ,ntov ,num, icord, rr, iel, del1,deltat,alpha,
     5idv4,sinv, icho, nnp, tref)
     
     
     
      implicit double precision(a-h,o-z)

      dimension node (maxel,maxst),pmat (maxel,   9),cord (maxnp, ndim)
      dimension ncod (maxdf      ),bc   (maxdf     ),sinv (maxel,   27)
      dimension vel  (maxnp ,   3),r1   (maxdf     ),conc (maxnp      )
      dimension aa   (maxst,maxst),rr   (maxst     ),ldest(maxst      )
      dimension xg   (          3),cg   (         3),kdest(maxst      )
      dimension x    (          3),v    (         3),nk   (maxst      )
      dimension bicn (          2),hh   (         3)
      dimension p    (         20),del  (   3,   20),b    (   3,    20)
      dimension eq   (maxfr,maxfr),nopp (maxdf     )     
      dimension ldsc (         22)
      dimension lhed (maxfr      ),khed (maxfr     ),jmod (maxfr      )
      dimension lpiv (maxfr      ),kpiv (maxfr     ),qq   (maxfr      )
      dimension pvkol(maxfr      ),mdf  (maxdf     ),ndn  (maxdf      )
      dimension ppp  (20    ,  20),pp   (20        )
	dimension ak   (100,100)
	dimension akf  (100    )   
      dimension NQ   (20    ,  20),NP   (3    ,   4)
	dimension C    (maxst      ),temp (maxnp     )
	dimension DEL1 (3          )
      dimension press(maxnp      ),clump(maxnp     ),SHAPE1D(3       )
	dimension gdsf (    3,   20)
	dimension dmass(100, 100)
	dimension aa01 (maxst,maxst)
	dimension aa02 (maxst,maxst) 
	dimension ak01 (maxst,maxst)
	dimension ak02 (maxst,maxst) 


           
      
                                                 rvisc = pmat(iel,1)
                                                 rpef  = pmat(iel,2)
                                                 power = pmat(iel,3)
                                                 rtem  = pmat(iel,4)
                                                 tbco  = pmat(iel,5)
                                                 taco  = pmat(iel,6)
                                                 roden = pmat(iel,8)
											     gamad = pmat(iel,9) 
                                          	    
          
      velsound = 6000.0
	beta     =    0.0
      
      
      do 6010 idf= 1,ntrix
                 rr (idf)      = 0.0
                 akf(idf)      = 0.0 
	           C  (idf)      = 0.0
			   do 6010 jdf= 1,ntrix

                 aa   (idf,jdf)=0.0
	           dmass(idf,jdf)=0.0
                 ak   (idf,jdf)=0.0
                 aa01 (idf,jdf) = 0.0
                 aa02 (idf,jdf)= 0.0
                 ak01 (idf,jdf)= 0.0
                 ak02 (idf,jdf)= 0.0
6010             continue
      
      
	if (ncn==4) then 
	call gausspt(ngaus,xg,cg,ncn)
	else if (ncn==8) then                               
      call gaussp(ngaus,xg,cg,ncn)
	end if

	       lg=0
      do 6020 ig=1,ngaus
                                    g = xg(ig)
      do 6020 jg=1,ngaus
                                    h = xg(jg)
      do 6020 kg=1,ngaus
                                    f = xg(kg)
                                                                
						            lg = lg + 1
 
  
      if(iter.eq.1) then


      call shape (g,h,f,p,del,ncn)

      call deriv (iel,ig,jg,kg,p,del,b,ncn,da,cg,node,cord,
     1            maxel,maxnp)

	              iig=ig
	              jjg=jg
                    kkg=kg

                    write(15)   iel ,ig ,jg ,kg, p ,del ,b ,da
      else
      if(.not. EOF(15))read(15) iel,iig,jjg,kkg,p ,del ,b , da

      endif
c
c     calculation of viscosity based on the constitutive equation.
c
      spress = 0.0
      stemp  = 0.0

	          do 5333 ip = 1,ncn
                jp    = iabs(node(iel,ip))
	          stemp = stemp + temp(jp) * p(ip)
 5333           continue
                epsii = 1.d-10
                gamad = sinv(iel,lg)
                if(gamad.lt.epsii) gamad = epsii
	
      call visca
	1(rvisc,power,visc,stemp,rtem,tbco,spress,rpef,taco,gamad)             

                       do 6050 idff= 1,3
                       x(idff)     = 0.0
                       v(idff)     = 0.0
                       hh(idff)    = 0.0
6050                   continue
          do 6060 icn = 1 ,ncn
                  jcn = iabs(node(iel,icn))
          do 6060 idff= 1 ,  3
          x(idff)     = x(idff) + p(icn)*cord(jcn,idff)
          v(idff)     = v(idff) + p(icn)*vel (jcn,idff)
6060		continue


      if(icord.eq.1) then
c
c     modify da for axisymmetric computations.
c
      da = da * x(1)
      endif

     
c     column index



      do 6070 i=1,ncn

       			   j11= i
                     j12= i + ncn
	               j13= i + 2*ncn
                     j14= i + 3*ncn
 
      do 6070 j=1,ncn
                     j21= j
                     j22= j + ncn
	               j23= j + 2*ncn
	               j24= j + 3*ncn
	               
	               
c	if (iel.le.3000) then               


c     Dicretized form of 3D Stokes Equation
c
c --- Stiffness Matrix of Left Hand Side -------------------------------------------------
c
c     For Transient state (Cartesian co-ordinate system) 

	
      aa(j11,j21)=aa(j11,j21) +  p(i)*p(j)*da
	1                        +  (alpha*deltat*2.0*(visc/roden)
	2						  +  0.5*alpha*alpha*deltat*deltat*velsound*velsound)
	3						  *  (b(1,i)*b(1,j)*da)
     3                        +  alpha*deltat*(visc/roden)
     4                        * ((b(2,i)*b(2,j)+b(3,i)*b(3,j))*da)   
     				
                
      aa(j11,j22)=aa(j11,j22) +  alpha*deltat*(visc/roden)
	1						  *  (b(2,i)*b(1,j)*da)
	2                        +  0.5*alpha*alpha*deltat*deltat*velsound
	3						  *  velsound*(b(1,i)*b(2,j)*da)
	
      aa(j11,j23)=aa(j11,j23) +  alpha*deltat*(visc/roden)
	1						  *  (b(3,i)*b(1,j)*da)
	2                        +  0.5*alpha*alpha*deltat*deltat*velsound
	3						  *  velsound*(b(1,i)*b(3,j)*da)
	
      aa(j11,j24)=aa(j11,j24) +  alpha*deltat*(b(1,j)
	1                           *p(i)*da)
	   
c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                       
      aa(j12,j21)=aa(j12,j21) +  alpha*deltat*(visc/roden)
	1						  *  (b(1,i) *b(2,j)*da)
     2                        +  0.5*alpha*alpha*deltat*deltat*velsound
	3						  *  velsound*(b(2,i)*b(1,j)*da)
         
      aa(j12,j22)=aa(j12,j22) +  p(i)*p(j)*da
	1                        +  (alpha*deltat*2.0*(visc/roden)
	2						  +  0.5*alpha*alpha*deltat*deltat*velsound*velsound)
	3						  *  (b(2,i)*b(2,j)*da)
     3                        +  alpha*deltat*(visc/roden)
     4                        * ((b(1,i)*b(1,j)+b(3,i)*b(3,j))*da) 					
     

      aa(j12,j23)=aa(j12,j23) +  alpha*deltat*(visc/roden)
	1						  *  (b(3,i)*b(2,j)*da)
	2                        +  0.5*alpha*alpha*deltat*deltat*velsound
	3						  *  velsound*(b(2,i)*b(3,j)*da)

      aa(j12,j24)=aa(j12,j24) +  alpha*deltat*(b(2,j)
	1                           *p(i)*da)
     	
c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

      aa(j13,j21)=aa(j13,j21) +  alpha*deltat*(visc/roden)
	1						  *  (b(1,i) *b(3,j)*da)
     2                        +  0.5*alpha*alpha*deltat*deltat*velsound
	3						  *  velsound*(b(3,i)*b(1,j)*da)
		

      aa(j13,j22)=aa(j13,j22) +  alpha*deltat*(visc/roden)
	1						  *  (b(2,i) *b(3,j)*da)
     2                        +  0.5*alpha*alpha*deltat*deltat*velsound
	3						  *  velsound*(b(3,i)*b(2,j)*da)

      aa(j13,j23)=aa(j13,j23) +  p(i)*p(j)*da
	1                        +  (alpha*deltat*2.0*(visc/roden)
	2						  +  0.5*alpha*alpha*deltat*deltat*velsound*velsound)
	3						  *  (b(3,i)*b(3,j)*da)
     3                        +  alpha*deltat*(visc/roden)
     4                        * ((b(1,i)*b(1,j)+b(2,i)*b(2,j))*da) 				


      aa(j13,j24)=aa(j13,j24) +  alpha*deltat*(b(3,j)
	1                           *p(i)*da)

c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

      aa(j14,j21)=aa(j14,j21) +  alpha*deltat*velsound*velsound
	1                        *  (p(i)*b(1,j)*da)


      aa(j14,j22)=aa(j14,j22) +  alpha*deltat*velsound*velsound
	1                        *  (p(i)*b(2,j)*da)
	

      aa(j14,j23)=aa(j14,j23) +  alpha*deltat*velsound*velsound
	1                        *  (p(i)*b(3,j)*da)


      aa(j14,j24)=aa(j14,j24) +  p(i)*p(j)*da
	1                        +  (0.5*alpha*alpha*deltat*deltat
	2                        *  velsound*velsound)*((b(1,i)*b(1,j)
     3                        +  b(2,i)*b(2,j)+b(3,i)*b(3,j))*da)
  
c --- Matrix on Right Hand Side -------------------------------------------------
c
c     For Transient State (Cartesian co-ordinate system) 


      ak(j11,j21)=ak(j11,j21) +  p(i)*p(j)*da
	1                        -  ((1.0-alpha)*deltat*2.0*(visc/roden)
	2						  +  0.5*alpha*(1.0-alpha)*deltat*deltat*velsound*velsound)
	3						  *  (b(1,i)*b(1,j)*da)
     3                        -  (1.0-alpha)*deltat*(visc/roden)
     4                        * ((b(2,i)*b(2,j)+b(3,i)*b(3,j))*da)  	
    
                  
      ak(j11,j22)=ak(j11,j22) -  (1.0-alpha)*deltat*(visc/roden)
	1						  *  (b(2,i)*b(1,j)*da)
	2                       -  0.5*alpha*(1.0-alpha)*deltat*deltat*velsound
	3						  *  velsound*(b(1,i)*b(2,j)*da)
		
      ak(j11,j23)=ak(j11,j23) -  (1.0-alpha)*deltat*(visc/roden)
	1						  *  (b(3,i)*b(1,j)*da)
	2                       -  0.5*alpha*(1.0-alpha)*deltat*deltat*velsound
	3						  *  velsound*(b(1,i)*b(3,j)*da)

	
      ak(j11,j24)=ak(j11,j24) -  (1.0-alpha)*deltat*(b(1,j)
	1                           *p(i)*da)
	   
c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
					                        
      ak(j12,j21)=ak(j12,j21) -  (1.0-alpha)*deltat*(visc/roden)
	1						  *  (b(1,i) *b(2,j)*da)
     2                        -  0.5*alpha*(1.0-alpha)*deltat*deltat
	3						  *  velsound*velsound*(b(2,i)*b(1,j)*da)


      ak(j12,j22)=ak(j12,j22) +  p(i)*p(j)*da
	1                        -  ((1.0-alpha)*deltat*2.0*(visc/roden)
	2						  +  0.5*alpha*(1.0-alpha)*deltat*deltat*velsound*velsound)
	3						  *  (b(2,i)*b(2,j)*da)
     3                        -  (1.0-alpha)*deltat*(visc/roden)
     4                        * ((b(1,i)*b(1,j)+b(3,i)*b(3,j))*da)   		


      ak(j12,j23)=ak(j12,j23) -  (1.0-alpha)*deltat*(visc/roden)
	1						  *  (b(3,i)*b(2,j)*da)
	2                       -  0.5*alpha*(1.0-alpha)*deltat*deltat*velsound
	3						  *  velsound*(b(2,i)*b(3,j)*da)
					   

      ak(j12,j24)=ak(j12,j24) -  (1.0-alpha)*deltat*(b(2,j)
	1                           *p(i)*da)

c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     	         
      ak(j13,j21)=ak(j13,j21) -  (1.0-alpha)*deltat*(visc/roden)
	1						  *  (b(1,i) *b(3,j)*da)
     2                        -  0.5*alpha*(1.0-alpha)*deltat*deltat
	3						  *  velsound*velsound*(b(3,i)*b(1,j)*da)

      ak(j13,j22)=ak(j13,j22) -  (1.0-alpha)*deltat*(visc/roden)
	1						  *  (b(2,i) *b(3,j)*da)
     2                        -  0.5*alpha*(1.0-alpha)*deltat*deltat
	3						  *  velsound*velsound*(b(3,i)*b(2,j)*da)



      ak(j13,j23)=ak(j13,j23) +  p(i)*p(j)*da
	1                        -  ((1.0-alpha)*deltat*2.0*(visc/roden)
	2						  +  0.5*alpha*(1.0-alpha)*deltat*deltat*velsound*velsound)
	3						  *  (b(3,i)*b(3,j)*da)
     3                        -  (1.0-alpha)*deltat*(visc/roden)
     4                        * ((b(1,i)*b(1,j)+b(2,i)*b(2,j))*da)  				


      ak(j13,j24)=ak(j13,j24) -  (1.0-alpha)*deltat*(b(3,j)
	1                           *p(i)*da)

c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

      ak(j14,j21)=ak(j14,j21) -  (1.0-alpha)*deltat*velsound*velsound
	1                        *  (p(i)*b(1,j)*da)



      ak(j14,j22)=ak(j14,j22) -  (1.0-alpha)*deltat*velsound*velsound
	1                        *  (p(i)*b(2,j)*da)

	
      ak(j14,j23)=ak(j14,j23) -  (1.0-alpha)*deltat*velsound*velsound
	1                        *  (p(i)*b(3,j)*da)
	

      ak(j14,j24)=ak(j13,j24) +  p(i)*p(j)*da
	1                        -  (0.5*alpha*(1.0-alpha)*deltat*deltat
	2                        *  velsound*velsound)*((b(1,i)*b(1,j)
     3                        +  b(2,i)*b(2,j)+b(3,i)*b(3,j))*da)


c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
c      end if

c
c     Body Force Effect (for Elemental Load Vector Calculation)....
c     ==========================================================
c
	C(j11) =C(j11) + (1.0-alpha)*deltat*p(j)*grav1*da     
    
      C(j12) =C(j12) + (1.0-alpha)*deltat*p(j)*grav2*da  

      C(j13) =C(j13) + (1.0-alpha)*deltat*p(j)*grav3*da  

      C(j14) =C(j14) + 0


      

6070  continue
      
6020  continue


c     Writing the Stiffness Matrix for Element Number 14

         If (iel==14) then
	
	write (idv4,3045) 
	write (idv4,3050) iter
      write (idv4,3055) ((aa(i,j),j=1,ntrix),i=1,ntrix)
         
3045  format(///,' ',10('*'),' element stiffness matrix for element 
     114',10('*'),///)
3050  format(///,' ','Iteration number =',i5,///) 
3055  format(27(E15.8,3x))
      
         end if 
         
         
 

c     For Transient State (Cartesian Co-ordinate System)
c     ================================================== 
c     Term one on RHS is evaluated 
c     ============================

      do 6080 i=1,ncn

       			   j11= i
                     j12= i + ncn
	               j13= i + 2*ncn
	               j14= i + 3*ncn
 
      do 6080 j=1,ncn
                     j21= j
                     j22= j + ncn
	               j23= j + 2*ncn
	               j24= j + 3*ncn

      nn=iabs(node(iel,j))


	akf(j11)=akf(j11) + ak(j11,j21)*vel(nn,1) + 
     1                    ak(j11,j22)*vel(nn,2) +
     2                    ak(j11,j23)*vel(nn,3) +
     3                    ak(j11,j24)*press(nn)
	
	
	akf(j12)=akf(j12) + ak(j12,j21)*vel(nn,1) + 
     1	                ak(j12,j22)*vel(nn,2) +
     2                    ak(j12,j23)*vel(nn,3) +
     3                    ak(j12,j24)*press(nn) 


	akf(j13)=akf(j13) + ak(j13,j21)*vel(nn,1) +
     1	                ak(j13,j22)*vel(nn,2) +
     2                    ak(j13,j23)*vel(nn,3) +
     3                    ak(j13,j24)*press(nn) 


	akf(j14)=akf(j14) + ak(j14,j21)*vel(nn,1) +
     1	                ak(j14,j22)*vel(nn,2) +
     2	                ak(j14,j23)*vel(nn,3) +
     3                    ak(j14,j24)*press(nn)
	
6080  continue
     
c
c    Evaluation of Elemental Load Vector
c    ===================================
   
      do 6085 i=1,ncn

       			   j11= i
                     j12= i + ncn
	               j13= i + 2*ncn
	               j14= i + 3*ncn
      
c     For Transient State  (Cartesian Co-ordinate System) 
      
	rr(j11)= rr(j11) + akf(j11) + C(j11)
      rr(j12)= rr(j12) + akf(j12) + C(j12)
	rr(j13)= rr(j13) + akf(j13) + C(j13)
	rr(j14)= rr(j14) + akf(j14) + C(j14)
 
6085  continue 

      
c
      maxte=maxdf
      call front
     1(aa   ,rr   ,iel  ,node ,maxel,maxst,ldest,kdest,nk   ,maxfr
     2,eq   ,lhed ,khed ,kpiv ,lpiv ,jmod ,qq   ,pvkol,vel  ,r1
     3,ncod ,bc   ,nopp ,mdf  ,ndn  ,maxdf,nel  ,maxte,ntov ,lcol
     4,nell ,ntra, press,icho )
c
      return
      end
               
c     ==================================================================
c     Stokes Solution based on the modified UVW-P scheme via 
c     Taylor-Galerkin time stepping method.
c     ==================================================================

      subroutine flowStokes2(node ,cord ,pmat ,nopp ,mdf ,ndn ,ncod ,
	1bc ,vel  ,press, r1, temp,ldest,kdest,nk ,eq ,lhed ,khed ,kpiv ,lpiv,
     2jmod, qq, pvkol, iter ,nel ,ncn , ngaus,grav1,
     3grav2, grav3, p, del, b, ntrix, maxel, maxnp, maxst, maxfr, maxdf,
     4ndim ,aa ,xg ,da ,ntov ,num, icord, rr, iel, del1,deltat,alpha,
     5idv4,sinv, icho, nnp, tref)
     
     
     
      implicit double precision(a-h,o-z)

      dimension node (maxel,maxst),pmat (maxel,   9),cord (maxnp, ndim)
      dimension ncod (maxdf      ),bc   (maxdf     ),sinv (maxel,   27)
      dimension vel  (maxnp ,   3),r1   (maxdf     ),conc (maxnp      )
      dimension aa   (maxst,maxst),rr   (maxst     ),ldest(maxst      )
      dimension xg   (          3),cg   (         3),kdest(maxst      )
      dimension x    (          3),v    (         3),nk   (maxst      )
      dimension bicn (          2),hh   (         3)
      dimension p    (         20),del  (   3,   20),b    (   3,    20)
      dimension eq   (maxfr,maxfr),nopp (maxdf     )     
      dimension ldsc (         22)
      dimension lhed (maxfr      ),khed (maxfr     ),jmod (maxfr      )
      dimension lpiv (maxfr      ),kpiv (maxfr     ),qq   (maxfr      )
      dimension pvkol(maxfr      ),mdf  (maxdf     ),ndn  (maxdf      )
      dimension ppp  (20    ,  20),pp   (20        )
	dimension ak   (100,100)
	dimension akf  (100    )   
      dimension NQ   (20    ,  20),NP   (3    ,   4)
	dimension C    (maxst      ),temp (maxnp     )
	dimension DEL1 (3          )
      dimension press(maxnp      ),clump(maxnp     ),SHAPE1D(3       )
	dimension gdsf (    3,   20)
	dimension dmass(100, 100)
	dimension aa01 (maxst,maxst)
	dimension aa02 (maxst,maxst) 
	dimension ak01 (maxst,maxst)
	dimension ak02 (maxst,maxst) 


           
      rvisc = pmat(iel,1)
      rpef  = pmat(iel,2)
      power = pmat(iel,3)
      rtem  = pmat(iel,4)
      tbco  = pmat(iel,5)
      taco  = pmat(iel,6)
      roden = pmat(iel,8)
	gamad = pmat(iel,9) 
                                          	    
          
      velsound = 1500.0
	beta     =    0.0
          
       lambda   =  10E2               
c      lambda   = (10E7-10E8)/visc    ! From Zienkienwicz 
c
      
      
      do 6010 idf= 1,ntrix
                 rr (idf)      = 0.0
                 akf(idf)      = 0.0 
	           C  (idf)      = 0.0
			   do 6010 jdf= 1,ntrix

                 aa   (idf,jdf)=0.0
	           dmass(idf,jdf)=0.0
                 ak   (idf,jdf)=0.0
                 aa01 (idf,jdf) = 0.0
                 aa02 (idf,jdf)= 0.0
                 ak01 (idf,jdf)= 0.0
                 ak02 (idf,jdf)= 0.0
6010             continue
      
      
	if (ncn==4) then 
	call gausspt(ngaus,xg,cg,ncn)
	else if (ncn==8) then                               
      call gaussp(ngaus,xg,cg,ncn)
	end if

	       lg=0
      do 6020 ig=1,ngaus
                                    g = xg(ig)
      do 6020 jg=1,ngaus
                                    h = xg(jg)
      do 6020 kg=1,ngaus
                                    f = xg(kg)
                                    
                            
						         lg = lg + 1
 
  
      if(iter.eq.1) then


      call shape (g,h,f,p,del,ncn)

      call deriv (iel,ig,jg,kg,p,del,b,ncn,da,cg,node,cord,
     1            maxel,maxnp)

	              iig=ig
	              jjg=jg
                    kkg=kg

                    write(15)   iel ,ig ,jg ,kg, p ,del ,b ,da
      else
      if(.not. EOF(15))read(15) iel,iig,jjg,kkg,p ,del ,b , da

      endif
c
c     calculation of viscosity based on the constitutive equation.
c
      spress = 0.0
      stemp  = 0.0

	          do 5333 ip = 1,ncn
                jp    = iabs(node(iel,ip))
	          stemp = stemp + temp(jp) * p(ip)
5333           continue
               epsii = 1.d-10
                gamad = sinv(iel,lg)
               if(gamad.lt.epsii) gamad = epsii
	
      call visca
	1(rvisc,power,visc,stemp,rtem,tbco,spress,rpef,taco,gamad)             

c
c     preparation of the convective acceleration terms/balancing
c     dissipation is used
c
                       do 6050 idff= 1,3
                       x(idff)     = 0.0
                       v(idff)     = 0.0
                       hh(idff)    = 0.0
6050                   continue
          do 6060 icn = 1 ,ncn
                  jcn = iabs(node(iel,icn))
          do 6060 idff= 1 ,  3
          x(idff)     = x(idff) + p(icn)*cord(jcn,idff)
          v(idff)     = v(idff) + p(icn)*vel (jcn,idff)
6060		continue

      if(icord.eq.1) then
c
c     modify da for axisymmetric computations.
c
      da = da * x(1)
      endif

     
c     column index

      do 6070 i=1,ncn

       			   j11= i
                     j12= i + ncn
	               j13= i + 2*ncn
                     j14= i + 3*ncn
 
      do 6070 j=1,ncn
                     j21= j
                     j22= j + ncn
	               j23= j + 2*ncn
	               j24= j + 3*ncn
	               
	               
           


c     Dicretized form of 3D Stokes Equation
c
c --- Stiffness Matrix of Left Hand Side ------------------------------------
c
c     For Transient state (Cartesian co-ordinate system) 

	
      aa(j11,j21)=aa(j11,j21) +  p(i)*p(j)*da
	1                        +  (alpha*deltat*2.0*(visc/roden)
	2		  +  0.5*lambda*alpha*alpha*deltat*deltat*velsound*velsound)
	3						  *  (b(1,i)*b(1,j)*da)
     3                        +  alpha*deltat*(visc/roden)
     4                        * ((b(2,i)*b(2,j)+b(3,i)*b(3,j))*da)   
     				
                
      aa(j11,j22)=aa(j11,j22) +  alpha*deltat*(visc/roden)
	1						  *  (b(2,i)*b(1,j)*da)
	2                    +  0.5*lambda*alpha*alpha*deltat*deltat*velsound
	3						  *  velsound*(b(1,i)*b(2,j)*da)
	
      aa(j11,j23)=aa(j11,j23) +  alpha*deltat*(visc/roden)
	1						  *  (b(3,i)*b(1,j)*da)
	2                 +  0.5*lambda*alpha*alpha*deltat*deltat*velsound
	3						  *  velsound*(b(1,i)*b(3,j)*da)
	
      aa(j11,j24)=aa(j11,j24) +  alpha*deltat*(b(1,j)
	1                           *p(i)*da)
	   
c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                       
      aa(j12,j21)=aa(j12,j21) +  alpha*deltat*(visc/roden)
	1						  *  (b(1,i) *b(2,j)*da)
     2               +  0.5*lambda*alpha*alpha*deltat*deltat*velsound
	3						  *  velsound*(b(2,i)*b(1,j)*da)
         
      aa(j12,j22)=aa(j12,j22) +  p(i)*p(j)*da
	1                        +  (alpha*deltat*2.0*(visc/roden)
	2	  +  0.5*lambda*alpha*alpha*deltat*deltat*velsound*velsound)
	3						  *  (b(2,i)*b(2,j)*da)
     3                        +  alpha*deltat*(visc/roden)
     4                        * ((b(1,i)*b(1,j)+b(3,i)*b(3,j))*da) 					
     

      aa(j12,j23)=aa(j12,j23) +  alpha*deltat*(visc/roden)
	1						  *  (b(3,i)*b(2,j)*da)
	2           +  0.5*lambda*alpha*alpha*deltat*deltat*velsound
	3						  *  velsound*(b(2,i)*b(3,j)*da)

      aa(j12,j24)=aa(j12,j24) +  alpha*deltat*(b(2,j)
	1                           *p(i)*da)
     	
c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

      aa(j13,j21)=aa(j13,j21) +  alpha*deltat*(visc/roden)
	1						  *  (b(1,i) *b(3,j)*da)
     2             +  0.5*lambda*alpha*alpha*deltat*deltat*velsound
	3						  *  velsound*(b(3,i)*b(1,j)*da)
		

      aa(j13,j22)=aa(j13,j22) +  alpha*deltat*(visc/roden)
	1						  *  (b(2,i) *b(3,j)*da)
     2                +  0.5*lambda*alpha*alpha*deltat*deltat*velsound
	3						  *  velsound*(b(3,i)*b(2,j)*da)

      aa(j13,j23)=aa(j13,j23) +  p(i)*p(j)*da
	1                        +  (alpha*deltat*2.0*(visc/roden)
	2	+  0.5*lambda*alpha*alpha*deltat*deltat*velsound*velsound)
	3						  *  (b(3,i)*b(3,j)*da)
     3                        +  alpha*deltat*(visc/roden)
     4                        * ((b(1,i)*b(1,j)+b(2,i)*b(2,j))*da) 				


      aa(j13,j24)=aa(j13,j24) +  alpha*deltat*(b(3,j)
	1                           *p(i)*da)

c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

      aa(j14,j21)=aa(j14,j21) +  lambda*alpha*deltat*velsound*velsound
	1                        *  (p(i)*b(1,j)*da)


      aa(j14,j22)=aa(j14,j22) +  lambda*alpha*deltat*velsound*velsound
	1                        *  (p(i)*b(2,j)*da)
	

      aa(j14,j23)=aa(j14,j23) +  lambda*alpha*deltat*velsound*velsound
	1                        *  (p(i)*b(3,j)*da)


      aa(j14,j24)=aa(j14,j24) +  p(i)*p(j)*da
	1                       +  (0.5*lambda*alpha*alpha*deltat*deltat
	2                        *  velsound*velsound)*((b(1,i)*b(1,j)
     3                        +  b(2,i)*b(2,j)+b(3,i)*b(3,j))*da)
  
c --- Matrix on Right Hand Side --------------------------------------------
c
c     For Transient State (Cartesian co-ordinate system) 


      ak(j11,j21)=ak(j11,j21) +  p(i)*p(j)*da
	1                  -  ((1.0-alpha)*deltat*2.0*(visc/roden)
	2 +  0.5*lambda*alpha*(1.0-alpha)*deltat*deltat*velsound*velsound)
	3						  *  (b(1,i)*b(1,j)*da)
     3                        -  (1.0-alpha)*deltat*(visc/roden)
     4                        * ((b(2,i)*b(2,j)+b(3,i)*b(3,j))*da)  	
    
                  
      ak(j11,j22)=ak(j11,j22) -  (1.0-alpha)*deltat*(visc/roden)
	1						  *  (b(2,i)*b(1,j)*da)
	2       -  0.5*lambda*alpha*(1.0-alpha)*deltat*deltat*velsound
	3						  *  velsound*(b(1,i)*b(2,j)*da)
		
      ak(j11,j23)=ak(j11,j23) -  (1.0-alpha)*deltat*(visc/roden)
	1						  *  (b(3,i)*b(1,j)*da)
	2         -  0.5*lambda*alpha*(1.0-alpha)*deltat*deltat*velsound
	3						  *  velsound*(b(1,i)*b(3,j)*da)

	
      ak(j11,j24)=ak(j11,j24) -  (1.0-alpha)*deltat*(b(1,j)
	1                           *p(i)*da)
	   
c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
					                        
      ak(j12,j21)=ak(j12,j21) -  (1.0-alpha)*deltat*(visc/roden)
	1						  *  (b(1,i) *b(2,j)*da)
     2                   -  0.5*lambda*alpha*(1.0-alpha)*deltat*deltat
	3						  *  velsound*velsound*(b(2,i)*b(1,j)*da)


      ak(j12,j22)=ak(j12,j22) +  p(i)*p(j)*da
	1                        -  ((1.0-alpha)*deltat*2.0*(visc/roden)
	2	 +  0.5*lambda*alpha*(1.0-alpha)*deltat*deltat*velsound*velsound)
	3						  *  (b(2,i)*b(2,j)*da)
     3                        -  (1.0-alpha)*deltat*(visc/roden)
     4                        * ((b(1,i)*b(1,j)+b(3,i)*b(3,j))*da)   		


      ak(j12,j23)=ak(j12,j23) -  (1.0-alpha)*deltat*(visc/roden)
	1						  *  (b(3,i)*b(2,j)*da)
	2         -  0.5*lambda*alpha*(1.0-alpha)*deltat*deltat*velsound
	3						  *  velsound*(b(2,i)*b(3,j)*da)
					   

      ak(j12,j24)=ak(j12,j24) -  (1.0-alpha)*deltat*(b(2,j)
	1                           *p(i)*da)

c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     	         
      ak(j13,j21)=ak(j13,j21) -  (1.0-alpha)*deltat*(visc/roden)
	1						  *  (b(1,i) *b(3,j)*da)
     2                     -  0.5*lambda*alpha*(1.0-alpha)*deltat*deltat
	3						  *  velsound*velsound*(b(3,i)*b(1,j)*da)

      ak(j13,j22)=ak(j13,j22) -  (1.0-alpha)*deltat*(visc/roden)
	1						  *  (b(2,i) *b(3,j)*da)
     2                     -  0.5*lambda*alpha*(1.0-alpha)*deltat*deltat
	3						  *  velsound*velsound*(b(3,i)*b(2,j)*da)



      ak(j13,j23)=ak(j13,j23) +  p(i)*p(j)*da
	1                        -  ((1.0-alpha)*deltat*2.0*(visc/roden)
	2 +  0.5*lambda*alpha*(1.0-alpha)*deltat*deltat*velsound*velsound)
	3						  *  (b(3,i)*b(3,j)*da)
     3                        -  (1.0-alpha)*deltat*(visc/roden)
     4                        * ((b(1,i)*b(1,j)+b(2,i)*b(2,j))*da)  				


      ak(j13,j24)=ak(j13,j24) -  (1.0-alpha)*deltat*(b(3,j)
	1                           *p(i)*da)

c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

      ak(j14,j21)=ak(j14,j21) -  lambda*(1.0-alpha)*deltat*velsound
	1                        *velsound * (p(i)*b(1,j)*da)



      ak(j14,j22)=ak(j14,j22) -  lambda*(1.0-alpha)*deltat*velsound
	1                        *  velsound * (p(i)*b(2,j)*da)

	
      ak(j14,j23)=ak(j14,j23) -  lambda*(1.0-alpha)*deltat*velsound
	1                        *  velsound * (p(i)*b(3,j)*da)
	

      ak(j14,j24)=ak(j13,j24) +  p(i)*p(j)*da
	1                 -  (0.5*lambda*alpha*(1.0-alpha)*deltat*deltat
	2                        *  velsound*velsound)*((b(1,i)*b(1,j)
     3                        +  b(2,i)*b(2,j)+b(3,i)*b(3,j))*da)


c~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
c
c     Body Force Effect (for Elemental Load Vector Calculation)
c     ==========================================================
c
	C(j11) =C(j11) + (1.0-alpha)*deltat*p(j)*grav1*da     
    
      C(j12) =C(j12) + (1.0-alpha)*deltat*p(j)*grav2*da  

      C(j13) =C(j13) + (1.0-alpha)*deltat*p(j)*grav3*da  

      C(j14) =C(j14) + 0


      

6070  continue
      
6020  continue


c     Writing the Stiffness Matrix for Element Number 14

         If (iel==14) then
	
	write (idv4,3045) 
	write (idv4,3050) iter
      write (idv4,3055) ((aa(i,j),j=1,ntrix),i=1,ntrix)
         
3045  format(///,' ',10('*'),' element stiffness matrix for element 
     114',10('*'),///)
3050  format(///,' ','Iteration number =',i5,///) 
3055  format(27(E15.8,3x))
      
         end if 
         
         
 

c     For Transient State (Cartesian Co-ordinate System)
c     ================================================== 
c     Term one on RHS is evaluated 
c     ============================

      do 6080 i=1,ncn

       			   j11= i
                     j12= i + ncn
	               j13= i + 2*ncn
	               j14= i + 3*ncn
 
      do 6080 j=1,ncn
                     j21= j
                     j22= j + ncn
	               j23= j + 2*ncn
	               j24= j + 3*ncn

      nn=iabs(node(iel,j))


	akf(j11)=akf(j11) + ak(j11,j21)*vel(nn,1) + 
     1                    ak(j11,j22)*vel(nn,2) +
     2                    ak(j11,j23)*vel(nn,3) +
     3                    ak(j11,j24)*press(nn)
	
	
	akf(j12)=akf(j12) + ak(j12,j21)*vel(nn,1) + 
     1	                ak(j12,j22)*vel(nn,2) +
     2                    ak(j12,j23)*vel(nn,3) +
     3                    ak(j12,j24)*press(nn) 


	akf(j13)=akf(j13) + ak(j13,j21)*vel(nn,1) +
     1	                ak(j13,j22)*vel(nn,2) +
     2                    ak(j13,j23)*vel(nn,3) +
     3                    ak(j13,j24)*press(nn) 


	akf(j14)=akf(j14) + ak(j14,j21)*vel(nn,1) +
     1	                ak(j14,j22)*vel(nn,2) +
     2	                ak(j14,j23)*vel(nn,3) +
     3                    ak(j14,j24)*press(nn)
	
6080  continue
     
c
c    Evaluation of Elemental Load Vector
c    ===================================
   
      do 6085 i=1,ncn

       			   j11= i
                     j12= i + ncn
	               j13= i + 2*ncn
	               j14= i + 3*ncn
      
c     For Transient State  (Cartesian Co-ordinate System) hnt 
      
	rr(j11)= rr(j11) + akf(j11) + C(j11)
      rr(j12)= rr(j12) + akf(j12) + C(j12)
	rr(j13)= rr(j13) + akf(j13) + C(j13)
	rr(j14)= rr(j14) + akf(j14) + C(j14)
 
6085  continue 

      
c
      maxte=maxdf
      call front
     1(aa   ,rr   ,iel  ,node ,maxel,maxst,ldest,kdest,nk   ,maxfr
     2,eq   ,lhed ,khed ,kpiv ,lpiv ,jmod ,qq   ,pvkol,vel  ,r1
     3,ncod ,bc   ,nopp ,mdf  ,ndn  ,maxdf,nel  ,maxte,ntov ,lcol
     4,nell ,ntra, press,icho )
c
      return
      end


c     ==================================================================
c
      subroutine concn
     1    (node ,cord ,pmat ,nopp ,mdf  ,ndn  ,ncod ,bc   ,conc ,vel
     2    ,r1   ,xg   ,ndim ,da   ,ldest,kdest,nk   ,eq   ,lhed
     3    ,khed ,kpiv ,lpiv ,jmod ,qq   ,pvkol,iter ,nel  ,ncn  ,ntov
     4    ,icord,ngaus, p    ,del  ,b    ,ntrix,maxel,maxnp
     5    ,maxst,maxfr,maxdf,num  ,ijmo ,ae
     6    ,re   ,cg   ,iel, del1  )
c
      implicit double precision(a-h,o-z)
c
      dimension node (maxel,maxst),cord (maxnp, ndim)
      dimension ncod (maxdf      ),bc   (maxdf      )
      dimension conc (maxnp      ),r1   (maxdf      ),vel  (maxnp,    3)
      dimension ae   (maxst,maxst),re   (maxst      )
      dimension p    (         20),del  (   3,    20),b    (   3,    20)
      dimension x    (          3),v    (          3)
      dimension bicn (          3),bjcn (          3)
      dimension hh   (          3),pmat (maxel,    20)
      dimension xg   (          3),cg   (          3)
      dimension nopp (maxdf      ),ldest(maxst      ),kdest(maxst      )
      dimension eq   (maxfr,maxfr),nk   (maxst      ),lhed (maxfr      )
      dimension khed (maxfr      ),kpiv (maxfr      ),lpiv (maxfr      )
      dimension jmod (maxfr      ),qq   (maxfr      ),pvkol(maxfr      )
      dimension mdf  (maxdf      ),ndn  (maxdf      )
	dimension ppp  (20    ,    20),pp   (20          )
	dimension kae  (maxst,maxst),kaef (maxst      )
	dimension NQ   (4    ,    3),NP   (3    ,    4)
	dimension C    (maxst      )
      dimension DEL1 (3          )
c
c
                                                 rvisc = pmat(iel,1)
                                                 rbulk = pmat(iel,2)
                                                 power = pmat(iel,3)
                                                 rtem  = pmat(iel,4)
                                                 tco   = pmat(iel,5)
                                                 roden = pmat(iel,6)
											   dispc = pmat(iel,7)	
c
c     Basic element loop 
c     ==================    
                             do 6010 itrix  = 1 ,maxst
                             re(itrix)      = 0.0
                             do 6010 jtrix  = 1 ,maxst
                             ae(itrix,jtrix)= 0.0
6010                         continue
c
c     Numerical integration
c     =====================
c
      call gaussp(ngaus,xg,cg)
c
	       lg=0

      do 6020 ig=1,ngaus
                                    g = xg(ig)
      do 6020 jg=1,ngaus
                                    h = xg(jg)
      do 6020 kg=1,ngaus
                                    f = xg(kg)
             lg=lg+1
                                          if(iter.eq.1) then
c
      call shape (g,h,f,p,del,ncn)
c
      call deriv (iel,ig,jg,kg,p,del,b,ncn,da,cg,node,cord,
     1            maxel,maxnp)

      end if
c
c     Read shape functions and their cartesian derivatives data from
c     a work file
      
	read (15)   iiel,iig,jjg,kkg,p,del,b,da
      
      ijmo=ijmo+1
c
c     Coefficients evaluated at integration point
c
                              do 6040 idf = 1 , 3
                              x(idf)      = 0.0
                              v(idf)      = 0.0
6040                          continue
          do 6050 icn = 1 ,ncn
                  jcn = iabs(node(iel,icn))
          do 6050 idf = 1 , 2
               x(idf) = x(idf) + p(icn)*cord(jcn,idf)
               v(idf) = v(idf) + p(icn)*vel (jcn,idf)
6050      continue
c
      if(icord.eq.1) then
c
c    Modify da for axisymmetric computations.
c    ======================================= 
      da = da * x(1)

      endif

      do 6060 icn = 1 ,ncn

c     Row index
c     =========
	                                 ir = icn
      do 6060 jcn = 1 ,ncn
                         
c     Column index
c     ============
                                       ic = jcn

6060  continue
6020  continue

c     Term one on RHS is evaluated 
c     ============================

      do 6080 icn = 1 , ncn

       			   ir = icn
 
      do 6080 jcn = 1 , ncn
                     
				   ic = jcn

      
	KAEF(ir)=KAEF(ir) + kae(ir,ic)*conc(ic)
	
6080  continue

c     Term two on RHS i.e. Calculation of Line Integrals 	
c     ==================================================
 
      NQ(1,3)=2
      NQ(1,2)=5
      NQ(1,1)=1
      NQ(2,3)=3
      NQ(2,2)=6
      NQ(2,1)=2
      NQ(3,3)=4
      NQ(3,2)=7
      NQ(3,1)=3
      NQ(4,3)=1
      NQ(4,2)=8
      NQ(4,1)=4 
      
	do I=1,4
	   do J=1,3
	      NP(4-J,1)=node(iel,NQ(I,4-J)) 
	   end do 
      end do 

	do 6096 L=1,4
		call gaussp(ngaus,xg,cg) 
		
		      do 6096 ig=1,ngaus
					   g = xg(ig)
					   if(iter.eq.1) then
      
                call deriv (iel,ig,jg,kg,p,del,b,ncn,da,cg,node,cord,
     1          maxel,maxnp)
c                


                call lagsh1( g , P , del)
			  
	          end if 
			  
			  dsix=0.0
			  dsiy=0.0
			  

			  do I=1,ngaus
			     dsix=dsix+p(I)*cord(NP(I,L),1)
				 dsiy=dsiy+p(I)*cord(NP(I,L),2)
			  end do 
			    
				concx=0.0 
	            concy=0.0
                  
		      do I=1,ngaus
                  
	            concx   = concx  + DEL1(I)*conc(NP(I,L))
                  concy   = concy  + DEL1(I)*conc(NP(I,L))
                  concz   = concz  + DEL1(I)*conc(NP(I,L))
				u11		= u11	 + DEL1(I)*vel (NP(I,L),1)
                  u12		= u12	 + DEL1(I)*vel (NP(I,L),1)
                  u21		= u21	 + DEL1(I)*vel (NP(I,L),2)
                  u22		= u22	 + DEL1(I)*vel (NP(I,L),2)
			  	velx    = velx   + DEL1(I)*vel (NP(I,L),1)
				vely    = vely   + DEL1(I)*vel (NP(I,L),2)
			  
			  end do
			  	     

c	ellgth = sqrt(dsix**2 + dsiy**2+dsiz**2)
		
		dcell	=	dsix/ellgth
		dcelm	=	dsiy/ellgth
c		dceln   =       dsiz/ellgth
		dnx	s	=	dcelm
	    dny		=	-dcell
      dnz        =      dceln
	    ellgth  =   2.0*ellgth
		djacob	=   ellgth/2.0
		   
      
	do 6087 icn=1,ncn

       			   ir = icn
                     
 6087  continue
 6096  continue

c     Calculation of the Elemental Load Vector
c     ========================================
      
      do 6085 icn =1 , ncn

       			   ir = icn
                  
      re(ir)  = re(ir) + kaef(ir) + C(ir) 
      
 6085 continue 

	
c     Solve equations
c     ===============

      maxte=maxnp
      call front
     1(ae   ,re   ,iel   ,node ,maxel,maxst,ldest,kdest,nk   ,maxfr
     2,eq   ,lhed ,khed  ,kpiv ,lpiv ,jmod ,qq   ,pvkol, conc ,r1
     3,ncod ,bc   ,nopp  ,mdf  ,ndn  ,maxdf,nel  ,maxte,ntov ,lcol
     4,nell ,ntra ,press,icho,c,akf,ak)
c
c     End of basic element loop
c     =========================

       return
       end
c


c     ==================================================================
      subroutine stress
     1     (nel,nnp,ncn ,node ,p , b , da ,vel  ,maxnp, maxel,  maxst , 
     2      maxdf, stres, press, rvisc ,clump ,ngaus  ) 
c  
       implicit double precision(a-h,o-z)
c
c function
c --------
c     calculates stress components at integration points, 
c     Direct Approach (coupled scheme)
c
      dimension node (maxel,maxst)  ,  b    (  3,  20)
      dimension stres(maxnp,    6)  ,  press(maxnp   ) , p (  20    )
      dimension vel  (maxnp,    3) 
      dimension clump(maxnp      )
c
                                                             rewind 15
c
                                                  do 4990 inp   =1,maxnp
                                                  do 4990 icp   =1, 6
                                                  stres(inp,icp)= 0.0
 4990                                             continue              
c

      do 5000 iel = 1 ,nel
c
      do 6010 ig=1,ngaus
                                    
      do 6010 jg=1,ngaus
                                    
      do 6010 kg=1,ngaus
                                    
       if(.not. EOF(15))read(15) iiel,iig,jjg,kkg,p ,del ,b , da
	                                    
				u11 = 0.0
                  u12 = 0.0
                  u13 = 0.0
                  u21 = 0.0
                  u22 = 0.0
                  u23 = 0.0
                  u31 = 0.0
                  u32 = 0.0
                  u33 = 0.0
	          pres1 = 0.0

        do 6020 icn = 1 ,ncn
                jcn = iabs(node(iel,icn))
                  u11 = u11 + b(1,icn)*vel(jcn,1)
                  u12 = u12 + b(2,icn)*vel(jcn,1)
                  u13 = u13 + b(3,icn)*vel(jcn,1)
                  u21 = u21 + b(1,icn)*vel(jcn,2)
                  u22 = u22 + b(2,icn)*vel(jcn,2)
                  u23 = u23 + b(3,icn)*vel(jcn,2)
                  u31 = u31 + b(1,icn)*vel(jcn,3)
                  u32 = u32 + b(2,icn)*vel(jcn,3)
                  u33 = u33 + b(3,icn)*vel(jcn,3)
			  pres1 = pres1 + p(icn)*press(jcn)

 6020   continue

c
c
c    cartesian components of the stress tensor
c    ========================================= 
c    Shear Stress (Tau)
c    =========================================

      sd11 = 2.0   *rvisc *  u11
      sd22 = 2.0   *rvisc *  u22
	sd33 = 2.0   *rvisc *  u33
      sd12 = rvisc * (u12 +  u21)
      sd13 = rvisc * (u13 +  u31)
      sd23 = rvisc * (u23 +  u32)	 

c    ========================================= 
c    Normal Stress (Pi)
c    =========================================

      s11  =-pres1 + sd11
      s22  =-pres1 + sd22
      s33  =-pres1 + sd33
      s12  = sd12
      s13  = sd13
      s23  = sd23

c    ================================================================ 
c *** calculate stress at nodal points 
c    ================================================================ 


         do 6500 icn = 1 ,ncn
                                    jcn = iabs(node(iel,icn))

                    stres(jcn,1)= stres(jcn,1)
     1                          + p(icn) *s11  *da
                    stres(jcn,2)= stres(jcn,2)
     1                          + p(icn) *s22  *da
                    stres(jcn,3)= stres(jcn,3)
     1                          + p(icn) *s33  *da
                    stres(jcn,4)= stres(jcn,4)
     1                          + p(icn) *s12  *da
                    stres(jcn,5)= stres(jcn,5)
     1                          + p(icn) *s13  *da
                    stres(jcn,6)= stres(jcn,6)
     1                          + p(icn) *s23  *da

 6500               continue

c
 6010 continue
 
 5000 continue

      return
c
      end
c
c     ==================================================================

      subroutine lumpm
     1    (clump,nnp  ,maxnp,nel  ,ngaus,p   ,del  ,b   ,maxst,
     2     node ,maxel,ncn  )
c
      implicit double precision(a-h,o-z)
      dimension b    (    3,   20) ,del  (   3,   20) ,p   (  20)
      dimension clump(maxnp      )
      dimension node (maxel,maxst)
      dimension pp   (ncn  ,ncn  )
c
                                    do 5000 inp = 1 ,nnp
                                    clump  (inp)= 0.0
 5000                               continue
c
              rewind 15
c
      do 5010 iel = 1 ,nel

             do 5020            ig  = 1 ,ngaus
             do 5020            jg  = 1 ,ngaus
             do 5020            kg  = 1 ,ngaus
                  

	if(.not. EOF(15)) read (15) jel ,iig ,jjg ,kkg ,p ,del ,b ,da
 
                     do 5030    icn = 1 ,ncn
                                 ww = 0.0
                     do 5040    jcn = 1 ,ncn
                                 ww = ww + p(icn)*p(jcn)*da
 5040                continue
                     inp            = iabs(node(iel,icn))
                     clump(inp)     = clump(inp) + ww

 5030                continue

 5020        continue
 5010        continue

      return
      end
c
c     ==================================================================

      subroutine getnod (nnp  ,cord ,idv1 ,idv2 ,maxnp,ndim,icord)
c
      implicit double precision(a-h,o-z)
c
c     arguments
c     =========
c     nnp    total number of nodal points in the mesh
c     cord   array for nodal coordinates
c     idv1   input device id. 
c     idv2   output device id. 
c     ndim   see below
c
      dimension cord(maxnp, ndim)
c
      if (.NOT. EOF(51)) read (idv1,1010) 
	1(jnp ,(cord(jnp,idf),idf=1,3),jnp=1,nnp)
      if(icord.eq.0) write(idv2,3010)
      if(icord.eq.1) write(idv2,3020)
      write(idv2,3030) (jnp ,(cord(jnp,idf),idf=1,3) ,jnp=1,nnp)
c
      return
c
1010  format(i8,e20.12,e20.12,e20.12)
3010  format(' ',///' ',20('*'),' nodal coordinates ',20('*'),//
     1' ',(7x,'id.',13x,'x-coord',13x,'y-coord',13x,'z-coord',13x)/)
3020  format(' ',///' ',20('*'),' nodal coordinates ',20('*'),//
     1' ',2(7x,'id/',7x,'r-coord',7x,'z-coord',20x)/)
3030  format(' ',i10,10x,f10.6,10x,f10.6,10x,f10.6)
c
      end
c
c     =================================================================

      subroutine getelm (nel  ,ncn  ,node ,idv1 ,idv2 ,maxel)
c
      implicit double precision(a-h,o-z)
c
c     arguments
c     =========
c     ncn    number of nodes per element
c     node   array for element connectivity data
c     idv1   input device id. 
c     idv2   output device id. 
c     maxel  see below
c
      dimension node (maxel, ncn)
c
              do 6010 iel = 1 ,nel
6010  if (.not. eof(51))read (idv1,1010) iel ,(node(iel,icn),icn=1,ncn)
      print*, "nodal connectivity array read"
      write(idv2,3010)
               do 6020 jel = 1 ,nel
6020  write(idv2,3020) jel ,(node(jel,icn),icn=1,ncn)
c
      return
c

1010  format(21i7)
3010  format(' ',///,' ',20('*'),' element connectivity ',20('*'),//
     1' ',4x,'id.',7x,'n o d a l - p o i n t  e n t r i e s',/)
3020  format(21i7)
c
      end
c
c     ==================================================================

      subroutine getbcd (nbc  ,ibc  ,jbc  ,vbc
     1                  ,idv1 ,idv2 ,maxbc)
c
      implicit double precision(a-h,o-z)
c
c     arguments
c     =========
c     nbc    number of nodal constraint data
c     ibc    array for constrained nodal points
c     jbc    array for constrained degree of freedom
c     vbc    array for boundary values
c     udv1   input device id.
c     idv2   output device id. 
c     maxbc  see below
c
      dimension ibc  (maxbc) ,jbc  (maxbc),vbc  (maxbc)
c
      if (.not. eof(51))read (idv1,1010) (ibc(ind) ,jbc(ind) ,vbc(ind)
	1    ,ind=1,nbc)
	print*, "boundary conditions array read" 
      write(idv2,3010)
      write(idv2,3020) (ibc(ind) ,jbc(ind) ,vbc(ind) ,ind=1,nbc)
c
      return
c
1010  format(2i5,f10.4)
3010  format(' ',//  /,' ',20('*'),' nodal constraint ',20('*'),//
     1' ',(8x,'id.',7x,'dof',10x,'value',10x)/)
3020  format(5x,i5,5x,i5,f17.4)
c
      end
c
c     =================================================================

      subroutine putbcv
     1 (nnp  ,nbc ,ibc ,jbc ,vbc ,ncod ,bc ,maxbc,maxdf,maxel,maxst,
     2  node)
c
      implicit double precision(a-h,o-z)
c
c     arguments
c     =========
c     ncod    array for constraint switch defined for every d.o.f.
c     bc      array for storing contraint value
c     maxbc   see below
c     maxdf   see below
c
      dimension ibc  (maxbc) ,jbc  (maxbc) ,vbc  (maxbc)
      dimension ncod (maxdf) ,bc   (maxdf) ,node (maxel,maxst)
c
      do 6010 ind = 1 ,nbc
      if(jbc(ind)>4) goto 6010
      jnd         = ibc(ind)+(jbc(ind)-1)*nnp
      bc    (jnd) = vbc(ind)
      ncod  (jnd) =  1
6010  continue

c     specifying the stress free condition on node number 84
c     ======================================================    
c	iel=16
c	inp=24   
c	kc=iabs(node(iel,inp))
c	
      return
      end
c
c    ==================================================================

      subroutine putbcc
     1    (nbc  ,ibc  ,jbc  ,vbc  ,ncod ,bc   ,maxbc,maxdf)
c
      implicit double precision(a-h,o-z)
c
c     arguments
c     =========
c     arguments same as subroutine putbcv
c
      dimension ibc  (maxbc) ,jbc  (maxbc) ,vbc  (maxbc)
      dimension ncod (maxdf) ,bc   (maxdf)
c
      do 6010 ind = 1 ,nbc
      if(jbc(ind).eq.5) then
      jnd         = ibc(ind)
      bc     (jnd)= vbc(ind)
      ncod   (jnd)= 1
      endif
6010  continue
c
      return
      end
c
c    ==================================================================

      subroutine putbcs
     1    (nnp  ,nbc  ,ibc  ,jbc  ,vbc  ,ncod ,bc   ,maxbc,maxdf)
c
      implicit double precision(a-h,o-z)
c
c     arguments
c     =========
c     arguments same as subroutine putbcv
c
      dimension ibc  (maxbc) ,jbc  (maxbc) ,vbc  (maxbc)
      dimension ncod (maxdf) ,bc   (maxdf)
c
      do 6010 ind = 1 ,nbc
      if(jbc(ind).eq.6) then
      jnd         = ibc(ind)
      bc     (jnd)= vbc(ind)
      ncod   (jnd)= 1
      endif
6010  continue
c
      return
      end
c
c    ==================================================================
      subroutine clean
     1    (ncn  ,nel  ,ndf  ,node ,r1   ,maxel,maxst,maxdf,
     2     bc   ,ncod ,icho )
c
      implicit double precision(a-h,o-z)
c
c     arguments
c     =========
c     all arguments are defined elsewhere.
c
      dimension r1   (maxdf) ,node(maxel,maxst)
      dimension bc   (maxdf) ,ncod(maxdf      )
c
c     function
c     ========
c     cleans the used arrays and makes them ready for  solution
c
           do  6010 i   = 1,maxdf
                  r1(i) = 0.0
                  bc(i) = 0.0
                ncod(i) = 0
6010                   continue
                ntrix   = ndf *ncn
           do  6020 iel = 1,nel
           do  6020 inp = 1,ntrix
           node(iel,inp) = iabs(node(iel,inp))
6020                   continue
           if(icho.ne.1)then
           do 6030 iel = 1,nel
           write(11,3010)iel,(node(iel,j),j=1,ncn)
3010       format(10i5)  
6030       continue
c
           endif
c
      return
      end
c
c    ==================================================================
      subroutine setprm
     1     (nnp  ,nel  ,ncn  ,node ,ndf  ,maxel,maxst,ndn  ,ntrix,
     2      maxdf,ntov ,mdf  ,nopp )
c
      implicit double precision(a-h,o-z)
c
c     arguments
c     =========
c     all arguments are defined elsewhere.
c
      dimension node (maxel,maxst), ndn  (maxdf)
      dimension mdf  (maxdf      ), nopp (maxdf)
c
c     function
c     ========
c     Sets the location data for nodal degrees of freedom
c
      do 6010 iel = 1 ,nel
         ndn(iel) = ntrix
      do 6010 icn = 1 ,ncn
              kcn = node(iel,icn)
			jacn= icn+(ndf-3)*ncn
              lacn= kcn+(ndf-3)*nnp

			jbcn= icn+(ndf-2)*ncn
              lbcn= kcn+(ndf-2)*nnp

			jccn= icn+(ndf-1)*ncn
              lccn= kcn+(ndf-1)*nnp

      node(iel,jacn) = lacn
      node(iel,jbcn) = lbcn
      node(iel,jccn) = lccn

6010  continue

      do 6020 idf = 1,ntov
         mdf(idf) = 1
         nopp(idf) = idf
6020  continue
c
      return
      end
c
c     ==================================================================
      subroutine getmat (nel ,nmat,pmat, idv1, idv2,maxel,rtem, rpef)
c
      implicit double precision(a-h,o-z)
c
c     arguments
c     =========
c     nmat  number of materials
c     pmat  array for material constants for each element
c     idv1 input device id.
c     idv2  output devide id.
c     maxel see below
c
      dimension pmat (maxel,   9)
c
      write(idv2,3010)
c
      do 6010 imat = 1 ,nmat
      if (.NOT. EOF(51)) read(idv1,1010) rvisc, power, tref, tbco, taco,
	1                                   dispc, pref, roden, gamad 
      print*, "material properties read"
CCCC                ifrom = 1
CCCC                ito   = nel
                
			  if(rtem .eq.0.) rtem  = 0.001
                if(rpef .eq.0.) rpef  = 0.001
      
	do 6020 iel  = 1 ,nel
         pmat(iel,1) = rvisc
         pmat(iel,2) = pref
         pmat(iel,3) = power
         pmat(iel,4) = tref
         pmat(iel,5) = tbco
         pmat(iel,6) = taco
         pmat(iel,7) = dispc
	   pmat(iel,8) = roden
         pmat(iel,9) = gamad
         rtem        = tref
	   rpef        = pref 

c     roden    density
c     rvisc    mu nought; consistency coefficient
c     pref     reference pressure
c     power    power law index
c     tref     reference temperature
c     tbco     coefficient b in the power law model
c     taco     coefficient a in the power law model
c     dispc    dispersion coefficient
c     gamad    shear rate
c
6020  continue
c
      write(idv2,3020) imat ,ifrom ,ito ,rvisc ,power
      write(idv2,3030)
      write(idv2,3040) tref ,tbco, pref, taco
      write(idv2,3050)
      write(idv2,3060) dispc , roden , gamad

6010  continue
c
      return
c
1010  format(9d10.5)
c
3010  format(' ',//' ',35('*'),' material properties ',35('*'),//
     1     ' ',2x,'id.',5x,'eid.(from-to)',3x,'consistency co-efficient'
     2,5x,'power law index',/)
3020  format(' ',i3,i12,i4,5x,g15.5,15x,g15.5)
3030  format(/x,' reference temperature   coefficient b
     1   reference pressure  coefficient a '/)
3040  format(f16.3,f22.4,6x,g10.3,9x,g10.3)
3050  format(/x,
     1'Dispersion Coefficient     Density          Shear rate'/)
3060  format(g13.3,15x,g7.1,6x,g16.5)
c
      end
c
c     ==================================================================
c
      subroutine contol
     1(vel  ,conc ,iter ,ntov ,nnp  ,maxnp,maxdf,errov,errop,vet ,cet,
     2 pet, press)  
     
c
      implicit double precision(a-h,o-z)
c
      dimension vel  (maxdf),conc (maxnp), press(maxnp)
      dimension vet  (maxdf),cet  (maxnp), pet  (maxnp)

	errv  = 0.0
      torv  = 0.0
      errc  = 0.0
      torc  = 0.0
	errp  = 0.0
	torp  = 0.0
c
c    calculate difference between velocities in consecutive iterations
c    =================================================================
c
                     do 6010 icheck = 1,ntov
      if(iter.eq.1) vet(icheck) = 0.0
C061008                     errv = errv +
C061008     1          (vel(icheck)-vet(icheck)) * (vel(icheck)-vet(icheck))

       errv = errv + 
     1       vel(icheck)*vel(icheck)- 2*vel(icheck)*vet(icheck)
     2      +vet(icheck)*vet(icheck)       
       torv = torv + vel(icheck)*vel(icheck)
c
      vet(icheck) = vel(icheck)
c
6010                 continue
                     errov= errv/torv
c
c    calculate difference between concentrations in consecutive iterations
c    =====================================================================
                     do 6020 icheck = 1,nnp
      if(iter.eq.1)  cet(icheck) = 0.0
       errc = errc +
     1      (conc(icheck)*conc(icheck))-2*conc(icheck)*cet(icheck)
     2      + cet(icheck)*cet(icheck)
                     torc = torc + conc(icheck)*conc(icheck)

      cet(icheck) = conc(icheck)
c
6020                 continue
                     erroc= errc/torc
c
c     calculate difference between pressures in consecutive iterations
c     =================================================================

                     do 6030 icheck = 1,nnp
      if(iter.eq.1)  pet(icheck) = 0.0

c230908            if(iter.eq.1)  press(icheck) = 0.0
       errp = errp +
     1        press(icheck)*press(icheck)-2*press(icheck)*pet(icheck)
     2      + pet(icheck)*pet(icheck) 
       torp = torp + press(icheck)*press(icheck)
c
             pet(icheck) = press(icheck)
c
6030                 continue
                     errop= errp/torp

c
      return
      end

c     ==================================================================

      subroutine output
     1     (nnp  ,vel  ,press, maxdf,maxnp,icord, stres)
c
      implicit double precision(a-h,o-z)
c
c     arguments are already defined
c     =============================
      dimension vel(maxdf), press(maxnp), conc(maxnp)
	dimension stres(maxnp,    6)
      
	write(60,3010)

      if(icord.eq.0) write(60,3020)
      if(icord.eq.1) write(60,3030)
      
	do 6010 inp = 1,nnp
                         jnp = inp + nnp
					   knp = inp + (2*nnp)
      press(inp)=press(inp)
      write(60,3040)inp,vel(inp),vel(jnp),vel(knp),press(inp),
	1stres(inp,1)
6010  continue

      call minimax
	1( cmax  , pmax  , vel  , conc  ,  press ,  maxnp,  nnp  ,    nc,    
	2  np    , nm    , ncm  , nvxm  ,  nvym  ,  nvzm ,  
     3  nvxl  , nvyl  , nvzl , pmin  ,  cmin  ,
     4  vxmax , vxmin , vymax, vymin ,  vzmax ,  vzmin,  ndim , maxdf )
      
	write(60,3045)
	write(60,3050)nvxm,vel(nvxm),nvxl,vel(nvxl)
	
	write(60,3055)
	write(60,3060)nvym,vel(nnp+nvym),nvyl,vel(nnp+nvyl)
	
	write(60,3065)
	write(60,3070)nvzm,vel(2*nnp+nvzm),nvzl,vel(2*nnp+nvzl)
	
	write(60,3075)
	write(60,3080)np,press(np),nm,press(nm)
	 

3010  format(/' nodal velocities and  pressures  '/)
3020  format('  id.      ux         uy          uz     press  stress'/)
3030  format('  id.      ur         uz          uz             press'/)
3040  format(i5,3e13.4,e22.8,g15.5)

3045  format('node no.         max ux    node no.         min ux')
3050  format(i5,e22.8,i5,e22.8,/)
  
3055  format('node no.         max uy    node no.         min uy')
3060  format(i5,e22.8,i5,e22.8,/)

3065  format('node no.         max uz    node no.         min uz')
3070  format(i5,e22.8,i5,e22.8,/)

3075  format('node no.         max p     node no.         min p')
3080  format(i5,e22.8,i5,e22.8,/)

      return
      end
c
c    ==================================================================
c    The subroutine slip introduces the slip wall boundary conditions
c    ==================================================================
      
	subroutine slip (ldsc)

      implicit double precision(a-h,o-z)

      dimension ldsc  (22         )


C
c    Channel depth data for slip wall b.c. & friction c.
c    ===================================================
 
      read(50,1010)rfrct
1010  format(f10.0)
c
c    This loop is used for identifying upper slip layer
c    ==================================================
 
      i1=0
      do 6030 ids=1,nel-13,14
         do 6030  j=1,2
         i1=i1+1
         ldsc(i1)=ids+j-1
6030  continue
c
c    This loop is used for identifying lower slip layers
c    ===================================================

      i1=0
      do 6020 ids=13,nel-2,14
         do 6020 j=1,2
         i1=i1+1
         ldsc(i1)=ids+j-1
6020  continue
c
c
      return
      end

c    ==================================================================

      subroutine hgstvl ( cmax, pmax , conc  ,  press , maxnp,
     1                    nnp , nc   , np   , nm   , pmin      )

      implicit real*8 (a-h,o-z)
      dimension conc  ( maxnp )
      dimension press ( maxnp )
      cmax= conc(1)
      pmax= press (1)
      pmin= press (1)
      nc=1
      np=1
      nm=1
      do i=2,nnp
         cm= conc(i)
         pm= press (i)
         pi= press (i)
         if ( cm.gt.cmax ) then
            cmax=cm
            nc  =i
         endif
         if ( pm.gt.pmax ) then
            pmax=pm
            np  =i
         endif
         if ( pi.lt.pmin ) then
            pmin = pi
            nm   = i
         endif
      enddo
      return
      end
c
c    ==================================================================
c     write nodal outputs                        .
c    
      subroutine output2 ( nnp , vel  , conc , press , maxdf, maxnp,
     1                    time , nwr  , iter , errov , erroc, errop)
                                     
      implicit real*8 (a-h,o-z)
      dimension vel   ( maxdf) , conc  ( maxnp) , press ( maxnp)
      dimension nwr (10) ,  vr  (10) , pv(10) , cr (10)
      
	call hgstvl ( cmax, pmax , conc  ,  press , maxnp,
     1                    nnp , nc   , np   , nm   , pmin      )

      write ( 2 , 5111) iter,errov,erroc,errop

      write ( 2 , 5115 ) time
      write ( 2,  5112) cmax,nc
      write ( 2,  5113) pmax,np
      write ( 2,  5114) pmin,nm
      write ( 2 , 5120 )
      do 24  i=1,nnp
        vres= dsqrt (vel(i+i-1)**2+vel(i+i)**2)
        psee= press(i)

        write (2,5130) i,vel(i+i-1),vel(i+i),vres,conc(i),press(i)

 24   continue

      call stress      
     1     (nel,nnp,ncn ,node ,p , b , da ,vel  ,maxnp, maxel,  maxst , 
     2      maxdf, stres, press, rvisc ,clump ,ngaus  ) 
c  
c     write the stress components
c     ===========================
  
      write ( 2 , 5133 )
      write ( 2 , 5130 ) i,   sd11,  sd12,  sd22
        
c     writing of output results
c     =========================

         do k=1,3
            i      = nwr(k)
            vr (k) = dsqrt (vel(i+i-1)**2+vel(i+i)**2)
            pv(k)  = press(i)
            cr (k) = conc (i)
         enddo
         write ( 4 , 5125 ) time , (vr(i),pv(i),cr(i),i=1,3)
c......................................................................
 5111 format (1x,/,'-  solution after',i5,'  iteration(s) -',
     1        1x,/,'-  error oval ( velocity ) =',f20.9,
     2        1x,/,'-  error oval ( concentration ) =',f20.9,
     3        1x,/,'-  error oval ( pressure ) =',f20.9)
 5112 format (1x,'  maximum concentration = ',g20.5,'  at node =',i5)
 5113 format (1x,'  maximum pressure    = ',g20.5,'  at node =',i5)
 5114 format (1x,'  minimum pressure    = ',g20.5,'  at node =',i5)
 5115 format (1x,'solution at time = ',g20.5,/)
 5120 format (1x,//,' result ( node no. ,vx, vy, |v|, concentration,
     1 pressure)',/)
 5125 format (1x,e11.6,3(' |',3e12.4))
 5130 format (1x,i4,2x,5(d11.5,2x))
 5133 format (1x,//,1x,' sd11   , sd12  , sd22 ',/)
      return
      end
c
c    ==================================================================
c     This subroutine calculate the viscosity using the power law model
c  
      subroutine visca
     1          (rvisc,power,visc,stemp,rtem,tbco,spress,rpef,taco
     2          ,gamad )

      implicit double precision(a-h,o-z)
       
c230908        visc   = rvisc*(4.0*gamad**((power-1.0)*0.5))
c230908	1           *exp(-tbco*(stemp-rtem))

c110909        visc = rvisc*(4.0*gamad**((power-1.0)*0.5))

               visc = rvisc*(gamad**((power-1.0))) 
  
      return
      end
c
c    ==================================================================
  
      subroutine lagsh1 ( xi , shape1d , del1 )

      implicit real*8 (a-h,o-z)
      dimension shape1d(3), del1(3)
        shape1d(1)   = -0.5*xi*(1.0-xi)
        shape1d(2)   = (1.0+xi)*(1.0-xi)
        shape1d(3)   =  0.5*g*(1.0+xi)
        del1(1)  = -0.5+xi
        del1(2)  = -2.0*xi
        del1(3)  =  0.5+xi
      return
      end


c     ==================================================================                            .
    
      subroutine minimax
	1( cmax  , pmax  , vel  , conc  ,  press ,  maxnp,  nnp  ,    nc,    
	2  np    , nm    , ncm  , nvxm  ,  nvym  ,  nvzm ,  
     3  nvxl  , nvyl  , nvzl , pmin  ,  cmin  ,
     4  vxmax , vxmin , vymax, vymin ,  vzmax ,  vzmin,  ndim , maxdf )

      implicit real*8 (a-h,o-z)
      dimension conc  ( maxnp )  , vel  (maxdf)
      dimension press ( maxnp )

c
      vxmax	=	vel(1)
	vxmin	=	vel(1)
	vymax	=	vel(nnp+1)
	vymin	=	vel(nnp+1)
	vzmax	=	vel(2*nnp+1)
	vzmin	=	vel(2*nnp+1)
 
      pmax	=	press (1)
      pmin	=	press (1)
      
	
	nc		=	1
	ncm		=	1
      np		=	1
      nm		=	1
	nvxm		=	1
	nvym		=	1
	nvzm		=	1
	nvxl		=	1
	nvyl		=	1
	nvzl		=	1
	      
	do 6020 i=2,nnp
      
      pm		=	press (i)
      pi		=   press (i)
	vxmx    =   vel(i)
	vxmn    =   vel(i)
	vymx    =   vel(nnp+i)
	vymn    =   vel(nnp+i)
	vzmx    =   vel(2*nnp+i)
	vzmn    =   vel(2*nnp+i)
         
         if (   pm.gt.pmax  ) then
            pmax=pm
            np  =i
         endif
         if (   pi.lt.pmin  ) then
            pmin = pi
            nm   = i
         endif
         if ( vxmx.gt.vxmax ) then
            vxmax= vxmx
            nvxm  = i
         endif
         if ( vymx.gt.vymax ) then
            vymax= vymx
            nvym  = i
         endif
         if ( vzmx.gt.vzmax ) then
            vzmax= vzmx
            nvzm  = i
         endif

	   if ( vxmn.lt.vxmin ) then
            vxmin= vxmn
            nvxl  = i
         endif
         if ( vymn.lt.vymin ) then
            vymin= vymn
            nvyl  = i
         endif
         if ( vzmn.lt.vzmin ) then
            vzmin= vzmn
            nvzl  = i
         endif
 6020 continue
      
	return
      end

c
c     ============================================================= 
c
      subroutine secinv
     1   (nel  ,nnp  ,ncn  ,ngaus,node ,sinv ,cord ,p  ,b ,
     2    del  ,da   ,vel  ,maxnp,maxel,maxst,ndim ,icord,
     3    maxdf,num)
c
      implicit double precision(a-h,o-z)
c
c function
c --------
c     calculates the second invariant of rate of deformation
c     tensor at integration points.
c
      dimension vel  (maxnp, ndim) ,cord (maxnp, ndim)
c      dimension vel  (maxdf      ) ,cord (maxnp, ndim)
      dimension node (maxel,maxst) ,sinv (maxel,   27)
      dimension p    (    20     ) ,del  (    3,   20)
      dimension b    (    3,   20)
c
                                         rewind 15
                                     
      do 5000 iel= 1 , nel
              lg = 0
      do 5010 ig = 1 ,ngaus
      do 5010 jg = 1 ,ngaus
	do 5010 kg = 1 ,ngaus

              lg = lg+1
c
           read (15) iiel,iig,jjg,p,del,b,da
c
                  u11 = 0.0
                  u12 = 0.0
				u13 = 0.0
                  u21 = 0.0
                  u22 = 0.0
	            u23 = 0.0
	            u31 = 0.0
	            u32 = 0.0
	            u33 = 0.0

          do 5020 icn = 1 ,ncn
                                    jcn = iabs(node(iel,icn))
c                                    mcn = jcn + nnp
c								  kcn = jcn + (2*nnp)

c *** components of the rate of deformation tensor

                  u11 = u11 + b(1,icn)*vel(jcn,1)
                  u12 = u12 + b(2,icn)*vel(jcn,1)
				u13 = u13 + b(3,icn)*vel(jcn,1)
                  u21 = u21 + b(1,icn)*vel(jcn,2)
                  u22 = u22 + b(2,icn)*vel(jcn,2)
				u23 = u23 + b(3,icn)*vel(jcn,2)
				u31 = u31 + b(1,icn)*vel(jcn,3)
                  u32 = u32 + b(2,icn)*vel(jcn,3)
				u33 = u33 + b(3,icn)*vel(jcn,3)

 5020     continue
c
c *** second invariant of the rate of deformation tensor
c
      sinv(iel,lg)=0.125*((u11+u11)*(u11+u11)+
     1                    (u12+u21)*(u12+u21)+
	2					(u13+u31)*(u13+u31)+
     3                    (u21+u12)*(u21+u12)+
     4                    (u22+u22)*(u22+u22)+
	5					(u23+u32)*(u23+u32)+
	6					(u31+u13)*(u31+u13)+
     7                    (u32+u23)*(u32+u23)+
	8					(u33+u33)*(u33+u33))
     
 5010 continue
 5000 continue 

      return
      end

c==================================================================
	subroutine gausspt(ngaus,xg,cg,ncn)
c
      implicit double precision(a-h,o-z)
c
c     x(g)   specifies the coordinates of the Gauss points
c     c(g)   specifies the Gauss weights
c
      dimension xg(5,5),cg(3)
	Real:: al, be
	al=0.58541020
	be=0.13819660
	
	xg=0.0
	cg=0.0
	
	if (ngaus==1) then
	xg(1,1)= 1.0/4.0
	xg(1,2)= xg(1,1)
	xg(1,3)= xg(1,1)
	xg(1,4)= xg(1,1)
	cg(1)=1
	else if (ngauss==2)then
	xg(1,1)=al
	xg(1,2)=be
	xg(1,3)=be	
	xg(1,4)=be
	xg(2,1)=be
	xg(2,2)=al
	xg(2,3)=be
	xg(2,4)=be
	xg(3,1)=be
	xg(3,2)=be
	xg(3,3)=al
	xg(3,4)=be
	xg(4,1)=be
	xg(4,2)=be
	xg(4,3)=be
	xg(4,4)=al
	cg(1)=1.0/4.0
	cg(2)=cg(1)
	else if (ngauss==3) then
	xg(1,1)=1.0/4.0
	xg(1,2)=1.0/4.0
	xg(1,3)=1.0/4.0	
	xg(1,4)=1.0/4.0

	xg(2,1)=1.0/2.0
	xg(2,2)=1.0/6.0
	xg(2,3)=1.0/6.0
	xg(2,4)=1.0/6.0

	xg(3,1)=1.0/6.0
	xg(3,2)=1.0/2.0
	xg(3,3)=1.0/6.0
	xg(3,4)=1.0/6.0

	xg(4,1)=1.0/6.0
	xg(4,2)=1.0/6.0
	xg(4,3)=1.0/2.0
	xg(4,4)=1.0/6.0

	xg(5,1)=1.0/6.0
	xg(5,2)=1.0/6.0
	xg(5,3)=1.0/6.0
	xg(5,4)=1.0/2.0

	end if 
	cg(1)=-4.0/5.0
	cg(2)=9.0/20.0
      return
      end

c===========================================================================

	SUBROUTINE GFMFEM
	
	PARAMETER (maxel  = 250000   )
	PARAMETER (maxnp  = 50000    )
	PARAMETER (maxbc  = 25000    )
      PARAMETER (maxdf  = maxnp*4  )
	PARAMETER (maxst  = 80       )
	PARAMETER (maxfr  = 2000     )
      PARAMETER (ndim   = 3        )

      IMPLICIT double PRECISION(a-h,o-z)


      CHARACTER CH(150)*1,SF*4,CC*2,FNAME*30,VL*2,VM*2,VN*2,CD*4,CE*5
      CHARACTER C1*2,C2*2,CW*4
	CHARACTER filnam (80)
      DIMENSION NOD(27)
      DIMENSION cord(maxnp, ndim),node (maxel, 27)

      WRITE (*,130 )
130   FORMAT(1X,'Enter GFORM file name    ',$)

      READ (*,135) FNAME
135   FORMAT (A30)

      OPEN (UNIT=1,FILE=FNAME,FORM='FORMATTED')
      OPEN (UNIT=2,STATUS='SCRATCH',FORM='FORMATTED')
      OPEN (UNIT=3,STATUS='SCRATCH',FORM='FORMATTED')
      OPEN (UNIT=4,STATUS='SCRATCH',FORM='FORMATTED')
      OPEN (UNIT=5,FILE='MESH.FEM',FORM='FORMATTED')
      OPEN (UNIT=6,FILE='input.dat',FORM='FORMATTED')
      OPEN (UNIT=7,STATUS='SCRATCH',FORM='FORMATTED')
      OPEN (UNIT=8,STATUS='SCRATCH',FORM='FORMATTED')
      OPEN (UNIT=9,STATUS='SCRATCH',FORM='FORMATTED')
      OPEN (UNIT=10,STATUS='SCRATCH',FORM='FORMATTED')
      OPEN (UNIT=11,STATUS='SCRATCH',FORM='FORMATTED') 


      IOS=0
      NND=0
      NEM=0
      NNE=0
      NNV=0
      NVN=0
      NNP=0
      NBC=0
      NCODE1=1
      NCODE2=2
      NCODE3=3
      NCODEP=4
	VNUL=0

c====================================================================
c      TO SCALE THE GEOMETRY
c====================================================================
      
1111 	PRINT *," Enter the scale factor: any number between 0.5 and 2 "
	READ*, scale
	IF(scale < 0.5 .OR. scale > 2) THEN
	PRINT*, " Invalid scale factor, try again." 
	
	GOTO 1111  
	
	END IF
	
C=====================================================================
C     DYNAMIC MEMORY ALLOCATION
C=====================================================================

      DO WHILE ( IOS.EQ.0 )
         READ (1,'(150A)',ERR=300,END=300,IOSTAT=IOS) (CH(J),J=1,150)
         CC=CH(1)//CH(2)
         CD=CH(1)//CH(2)//CH(3)
         CE=CH(1)//CH(2)//CH(3)//CH(4)//CH(5)
	   
	   IF (CC.EQ.'ND')THEN
             WRITE (2,'(100A)') (CH(K),K=4,100)
             WRITE (8,'(100A)') (CH(K),K=4,100)
             NND=NND+1
         ENDIF
         
	   IF (CC.EQ.'EL')THEN
             WRITE (2,'(100A)') (CH(K),K=4,100)
             WRITE (8,'(100A)') (CH(K),K=4,100) 
             NEM=NEM+1
         ENDIF
     
                         
         IF (CD.EQ.'VND') THEN
             WRITE (3,'(50A)') (CH(K),K=5,30)
             WRITE (9,'(50A)') (CH(K),K=5,30)
             NNV=NNV+1
         ENDIF
	   IF (CE.EQ.'NPRND') THEN
             WRITE (4,'(150A)') (CH(K),K=7,85)
             WRITE (10,'(150A)') (CH(K),K=7,85)
             NNP=NNP+1
         ENDIF
         IF (CW.EQ.'ND')THEN
             WRITE (7,'(100A)') (CH(K),K=4,100)
             WRITE (11,'(100A)') (CH(K),K=4,100)
             NND=NND+1
         ENDIF
         IF (CW.EQ.'VX')THEN
             WRITE (7,'(100A)') (CH(K),K=4,100)
             WRITE (11,'(100A)') (CH(K),K=4,100)
             NND=NND+1
         ENDIF
         IF (CW.EQ.'VY')THEN
             WRITE (7,'(100A)') (CH(K),K=4,100)
             WRITE (11,'(100A)') (CH(K),K=4,100)
             NND=NND+1
         ENDIF
         IF (CW.EQ.'VZ')THEN
             WRITE (7,'(100A)') (CH(K),K=4,100)
             WRITE (11,'(100A)') (CH(K),K=4,100)
             NND=NND+1
         ENDIF 
         
      ENDDO
      
300   REWIND 2
      REWIND 3
      REWIND 4
      REWIND 7
      REWIND 8
      REWIND 9
      REWIND 10
      REWIND 11

      
C==================================================================
C       NODAL COORDINATES
C==================================================================
      
      DO I=1,NND
        READ (2,*) N,X,Y,Z
	      X=X/scale
	      Y=Y/scale
	      Z=Z/scale
	  WRITE (5,'(I8,3G20.8)') N,X,Y,Z
      ENDDO
      
C010908============================================================

       DO I=1,NND
            READ (7,*,ERR=333,END=333,IOSTAT=IOS) K,VX,VY,VZ,X,Y,Z
            WRITE (11) VX,VY,VZ,X,Y,Z
       ENDDO
333    REWIND(7)      

C==================================================================
C       ELEMENT CONNECTIVITY
C==================================================================
      
      DO I=1,NEM
        READ (2,*) NNEE,SF,NSF,NPE,(NOD(K),K=1,NPE)
      
      
	  WRITE (5,120) I,NOD(2),NOD(6),NOD(5),NOD(1),NOD(3),
	1                   NOD(7),NOD(8),NOD(4) 
      
120   FORMAT (9I8)
      ENDDO
      
C==================================================================
C       VELOCITY BOUNDARY CONDITION
C==================================================================
      
      DO J= 1,NNV
        READ (3,1601) ND,VL,CU,V,NT,NP,VM,VN
1601	FORMAT (I7,A2,A1,G20.8,2I5,A2,A3)

C--------------------------------------------------
        IF (NP.EQ.0) THEN 
!-----VND,ND,VX,VAL
            IF (VL.EQ.'VX') THEN 
      	        WRITE (5,'(2I5,F10.4)') ND,NCODE1,V
!-----VND,ND,VY,VAL
	      ELSE IF (VL.EQ.'VY') THEN
	          WRITE (5,'(2I5,F10.4)') ND,NCODE2,V
!-----VND,ND,VZ,VAL	
	      ELSE IF (VL.EQ.'VZ') THEN
	          WRITE (5,'(2I5,F10.4)') ND,NCODE3,V
	      END IF 
	          NVN=NVN+1
	          NBC=NBC+1
	  ELSE IF (NP.EQ.1) THEN
             IF (VL.EQ.'VX') THEN         
!------VND,ND,VX,VAL,ND,NPE,VY,VZ	 
	          IF((VM.EQ.'VY').AND.(VN.EQ.'VZ'))THEN
	              WRITE (5,'(2I5,F10.4)') ND,NCODE1,V
	              WRITE (5,'(2I5,F10.4)') ND,NCODE2,V
	              WRITE (5,'(2I5,F10.4)') ND,NCODE3,V
	              NVN = NVN + 3
	              NBC = NBC + 3
!------VND,ND,VX,VAL,ND,NPE,VY	
	          ELSE IF (VM.EQ.'VY') THEN 
	              WRITE (5,'(2I5,F10.4)') ND,NCODE1,V
	              WRITE (5,'(2I5,F10.4)') ND,NCODE2,V
	              NVN = NVN + 2
	              NBC = NBC + 2
!------VND,ND,VX,VAL,ND,NPE,VZ	
	          ELSE IF(VM.EQ.'VZ') THEN
	              WRITE (5,'(2I5,F10.4)') ND,NCODE1,V
	              WRITE (5,'(2I5,F10.4)') ND,NCODE3,V
	              NVN = NVN + 2
	              NBC = NBC + 2
	              
            ElSE IF(VL.EQ.'VX')Then
!-----VND,ND,VX,VAL,NPE
	      WRITE (5,'(2I5,F10.4)') ND,NCODE1,V
	      NVN=NVN+1
	      NBC=NBC+1
CCC	      END IF
            END IF
!-----VND,VY,VAL,ND,NPE,VZ
        ELSE IF(VL.EQ.'VY'.AND.(VM.EQ.'VZ'))THEN
	      WRITE (5,'(2I5,F10.4)') ND,NCODE2,V
	      WRITE (5,'(2I5,F10.4)') ND,NCODE3,V
	          NVN = NVN + 2
	          NBC = NBC + 2
	  ELSE IF(VL.EQ.'VY')then
	      WRITE (5,'(2I5,F10.4)') ND,NCODE2,V
                NVN=NVN+1
	          NBC=NBC+1
        ELSE IF(VL.EQ.'VZ') THEN
                WRITE (5,'(2I5,F10.4)') ND,NCODE3,V
		        NVN=NVN+1
                NBC=NBC+1
        END IF 
       End IF
      END DO
      
C==================================================================
C       PRESSURE BOUNDARY CONDITION
C==================================================================
       
        DO I=1,NNP
            READ (4,'(I5,G20.8)') PC,VP
            WRITE (5,'(2I5,F10.4)' ) PC,NCODEP,VP
            NBC=NBC+1
            NAT=NAT+1
        END DO
        
C==================================================================
C     Data file preparation
C==================================================================

      ncn=NPE
	ngauss = 3
	nmat=1
	ntep=1
	icord=0
      grav1=0.0 
	grav2=0.0
	grav3=0.0
	tolv= 1e-05
	tolp= 1e-05
	tolc= 1e-05
	rvisc = 80.0
      power = 1.0
	tref= 293.0
	tbco = 0.014
	taco = 0.2
	dispc = 0.2
	pref =1.01325e5
	roden = 1000.0
	
c==================================================================
c     writing the data file
c==================================================================
   
      write (6,*) FNAME
	write (6,'(2i5)') ncn,ngauss
	write (6,'(4i5)') NND,NEM,NBC,nmat
	write (6,'(2i5)') ntep,icord
	write (6,'(3f10.3)') grav1, grav2, grav3
	write (6,'(3f10.5)') tolv,tolp,tolc
	write (6,'(9d10.5)') rvisc, power, tref, tbco, taco,
	1               dispc, pref, roden, gamad

C==================================================================
C       NODAL COORDINATES
C==================================================================
      
      DO I=1,NND
        READ (8,*) N,X,Y,Z
	      X=X/scale
	      Y=Y/scale
	      Z=Z/scale
	      cord(I,1)=X
	      cord(I,2)=Y
	      cord(I,3)=Z
	      WRITE (6,'(I8,3e20.12)') N,X,Y,Z
      ENDDO

C==================================================================
C       ELEMENT CONNECTIVITY
C==================================================================
      
      DO I=1,NEM
        READ (8,*) NNEE,SF,NSF,NPE,(node(I,K),K=1,NPE)
     
            if (NPE==4) then
                WRITE (6,'(5I8)')I,node(I,3),node(I,2),
     1                 node(I,1),node(I,4)
            else if (npe==8) then
	          WRITE (6,'(21I7)') I,node(I,2),node(I,6),
	2                 node(I,5),node(I,1),node(I,3),
     3                 node(I,7),node(I,8),node(I,4)	
	      else if (NPE==20) then 
		        WRITE (6,'(10I8)') I,NOD(2),NOD(6),NOD(5),
	4                 NOD(1),NOD(3),NOD(7),NOD(8),NOD(4) 
            end if 
      ENDDO
      
C==================================================================
C       VELOCITY BOUNDARY CONDITION
C==================================================================
          
	DO J= 1,NNV

        READ (9,1602) ND,VL,CU,V,NT,NP,VM,VN
1602	FORMAT (I7,A2,A1,G20.8,2I5,A2,A3)

        IF (NP.EQ.0) THEN 
            IF (VL.EQ.'VX') THEN 
      	        WRITE (6,'(2I5,F10.4)') ND,NCODE1,V
	      ELSE IF (VL.EQ.'VY') THEN
	          WRITE (6,'(2I5,F10.4)') ND,NCODE2,V
	      ELSE IF (VL.EQ.'VZ') THEN
	          WRITE (6,'(2I5,F10.4)') ND,NCODE3,V
	      END IF 
	  ELSE IF (NP.EQ.1) THEN
            IF (VL.EQ.'VX') THEN 
	          IF((VM.EQ.'VY').AND.(VN.EQ.'VZ'))THEN
	              WRITE (6,'(2I5,F10.4)') ND,NCODE1,V
	              WRITE (6,'(2I5,F10.4)') ND,NCODE2,V
	              WRITE (6,'(2I5,F10.4)') ND,NCODE3,V
	          ELSE	IF (VM.EQ.'VY') THEN 
	              WRITE (6,'(2I5,F10.4)') ND,NCODE1,V
	              WRITE (6,'(2I5,F10.4)') ND,NCODE2,V
	          ELSE IF(VM.EQ.'VZ') THEN
	              WRITE (6,'(2I5,F10.4)') ND,NCODE1,V
	              WRITE (6,'(2I5,F10.4)') ND,NCODE3,V
                ElSE IF(VL.EQ.'VX')Then
	              WRITE (6,'(2I5,F10.4)') ND,NCODE1,V
	          END IF
            ELSE IF(VL.EQ.'VY'.AND.(VM.EQ.'VZ'))THEN
	          WRITE (6,'(2I5,F10.4)') ND,NCODE2,V
	          WRITE (6,'(2I5,F10.4)') ND,NCODE3,V
	      ELSE IF(VL.EQ.'VY')then
	          WRITE (6,'(2I5,F10.4)') ND,NCODE2,V
            ELSE IF(VL.EQ.'VZ') THEN
                WRITE (6,'(2I5,F10.4)') ND,NCODE3,V
            END IF 
         END IF
      END DO

C==================================================================
C       PRESSURE BOUNDARY CONDITION
C==================================================================
        
        DO I=1,NNP
            READ (10,'(I5,G20.8)') PC,VP
                WRITE (6,'(2I5,F10.4)' ) PC,NCODEP,VP
        END DO
        
C==================================================================
C       OUTPUT OF RESULTS
C==================================================================

         PRINT *,' '
         PRINT *,"The geometry is discretized into a finite element
     1 mesh of: "
         PRINT *,' '
         PRINT *,NND,
     1    "Nodes" 
         PRINT *,' ' 
         PRINT *,NEM,
     2    "Elements"
	   PRINT *,"With"
         PRINT *,NNP,
     4    'Applied nodal pressure boundary conditions'
	   PRINT *,"And"
         PRINT *, NVN,
     5   "Applied nodal velocity boundary conditions"
         PRINT *,"Giving a"
         PRINT *,NBC,
     3    'total number of applied boundary conditions' 
     
c         PRINT*,' '
c         PRINT*, ' THE SHEAR RATE = ',GAMAD
        
C=====================================================================

      CLOSE (1)
      CLOSE (2)
      CLOSE (3)
      CLOSE (4)
      CLOSE (5)
	CLOSE (6)
      CLOSE (11)
	print *,'       '
	print *,'       '
      return
200   END

c     ===================================================================
c     This subroutine prepares output data to use for visualization
c     using Cosmos GeoStar software.  
c     
C     ===================================================================
      subroutine cosmos
	1       (nnp  ,  vel  ,  press   ,  maxdf  ,  maxnp  , icord   ,
     2        pmat ,  maxel,  actpress ,nel                          )
c
      implicit double precision(a-h,o-z)
c
c     arguments are already defined
c     =============================
      dimension vel (maxdf    ), press   (maxnp) 
	dimension pmat(maxel,  9), actpress(maxnp)
	dimension vm      (nnp  )
	

     	open(unit=610   ,   file='cosmGraph',   access='sequential',
     1form='formatted',   status="unknown" ,   iostat=ios         )
      
	j=0
	k=5

	write(610,3010) nnp, j, k
      
c      do i = 1, nel
c        roden=pmat(1,8)
c      end do

      DO inp=1,nel
      roden=pmat(inp,8)
      END DO 
      
c      do i=1,nnp
c	j=i
c	vm(j)=sqrt((vel(j,1)**2)+(vel(j,2)**2)+(vel(j,3)**2))
c	end do 
      
      
      
        
      
      write (610,6000) 

      do 6010 inp = 1,nnp
                        jnp = inp + nnp
                        knp = inp + (2*nnp)
      
c	actpress(inp)=roden*press(inp)*-1

      actpress(inp)=roden*press(inp)
      vm(inp)=sqrt((vel(inp)**2)+(vel(jnp)**2)+(vel(knp)**2))
      write(610,3020)inp,vel(inp),vel(jnp),vel(knp),
     1               vm(inp),actpress(inp)

6010  continue

      close (610)

3010  format(3i5)
3020  format(i5,4e13.4,e22.8)
6000  FORMAT(8X,'U',8X,'V',5X,'W',5X, 'M'
     1       8X,'PRESSURE',/) 

      return
      end
      

c     ==================================================================


c     ==================================================================
c     This subroutine prepares output data to use for visualization
c     using tecplot software.  
c     ==================================================================


      subroutine tecplot
	1       (nnp  ,  vel  ,  press   ,  maxdf  ,  maxnp  , icord   ,
     2        pmat ,  maxel,  actpress,  cord   ,  ncn    , nel     ,
     3        node ,  ndim  )
c
      implicit double precision(a-h,o-z)
c
c     arguments are already defined
c     =============================
      dimension vel     (nnp,3), press (maxnp)     , pmat(maxel ,  9)
	dimension actpress(maxnp), cord  (maxnp,ndim), node(maxel, ncn)
	dimension vm      (nnp  )

     	open(unit=614   ,   file='tecpGraph.dat',   access='sequential',
     1form='formatted',   status="unknown" ,   iostat=ios         )
      
c     Compute the Magnitude of the resultant velocity
c     ===============================================	

      do i=1,nnp
	j=i
	vm(j)=sqrt((vel(j,1)**2)+(vel(j,2)**2)+(vel(j,3)**2))
	end do 
	

c     Write the Techplot file for post-processing
c     ===========================================
c      nnel=4*nel

      write (614,1000)
	write (614,2000) nnp, nel

      roden=pmat(1,8)

      do i=1,nnp
c	j=i
c	actpress(i)=roden*press(i)* -1
      actpress(i)=roden*press(i)
      
	write (614,5000) cord(i,1), cord(i,2), cord(i,3),
	1                 vel(i,1), vel(i,2), vel(i,3),
     2                 vm(i), actpress(i) 	
      end do 


c230908      do i=1,nel
c230908	j=i
c230908	write (614,6000) abs(node(j,1)), abs(node(j,2)), abs(node(j,9))
c230908	1                ,abs(node(j,8))
c230908      write (614,6000) abs(node(j,2)), abs(node(j,3)), abs(node(j,4))
c230908	1                ,abs(node(j,9))
c230908	write (614,6000) abs(node(j,9)), abs(node(j,4)), abs(node(j,5))
c230908	1                ,abs(node(j,6))
c230908	write (614,6000) abs(node(j,8)), abs(node(j,9)), abs(node(j,6))
c230908	1                ,abs(node(j,7))
c230908      end do 
      
      
c     Elemental connectivity for techplot files
c-------------------------------------------------

      do  i=1,nel
c	j=i
      if (ncn==8)then
	write (614,'(8i8)') abs(node(i,1)), abs(node(i,2)), abs(node(i,6))
	1                   ,abs(node(i,5)), abs(node(i,4)), abs(node(i,3))
     2                   ,abs(node(i,7)), abs(node(i,8))
      else if (ncn==4) then
      write (614,'(4i8)') abs(node(i,2)), abs(node(i,3)), abs(node(i,4))
     1                  , abs(node(i,1))
      end if

      end do  
      
      

      close (614)

1000  format(/'Variables = "X", "Y","Z","U","V","W","M","P"'/)
c 2000  format(/'ZONE N=',i5,',E=',i5,',F=FEPOINT,ET=QUADRILATERAL'/)


2000  format(/'ZONE N=',i5,',E=',i5,',F=FEPOINT,ET=BRICK'/)

5000  format(3e20.12,3e13.4,e13.4,e22.8) 
6000  format(4i8) 

      return
      end






c     ==================================================================
c     e    n    d    o    f    p    r    o    g    r    a    m 
c     ==================================================================

